""" Test case for issue #10626: nested builtins.min / builtins.max calls"""

import builtins

builtins.min(1, min(2, 3))  # [nested-min-max]
min(1, builtins.min(2, 3)) # [nested-min-max]
builtins.min(1, builtins.min(2, 3))  # [nested-min-max]

builtins.max(1, max(2, 3)) # [nested-min-max]
max(1, builtins.max(2, 3)) # [nested-min-max]
builtins.max(1, builtins.max(2, 3)) # [nested-min-max]



"""Check for nonlocal and used-before-assignment"""
# pylint: disable=missing-docstring, unused-variable, too-few-public-methods


def test_ok():
    """ uses nonlocal """
    cnt = 1
    def wrap():
        nonlocal cnt
        cnt = cnt + 1
    wrap()

def test_fail():
    """ doesn't use nonlocal """
    cnt = 1
    def wrap():
        cnt = cnt + 1 # [used-before-assignment]
    wrap()

def test_fail2():
    """ use nonlocal, but for other variable """
    cnt = 1
    count = 1
    def wrap():
        nonlocal count
        cnt = cnt + 1 # [used-before-assignment]
    wrap()

def test_fail3(arg: test_fail4): # <3.14:[used-before-assignment]
    """ Depends on `test_fail4`, in argument annotation. """
    return arg
# +1:<3.14: [used-before-assignment, used-before-assignment]
def test_fail4(*args: test_fail5, **kwargs: undefined):
    """ Depends on `test_fail5` and `undefined` in
    variable and named arguments annotations.
    """
    return args, kwargs

def test_fail5()->undefined1: # <3.14:[used-before-assignment]
    """ Depends on `undefined1` in function return annotation. """

def undefined():
    """ no op """

def undefined1():
    """ no op """


def nonlocal_in_ifexp():
    """bar"""
    bug2 = True
    def on_click(event):
        """on_click"""
        if event:
            nonlocal bug2
            bug2 = not bug2
    on_click(True)

nonlocal_in_ifexp()


def type_annotation_only_gets_value_via_nonlocal():
    """https://github.com/pylint-dev/pylint/issues/5394"""
    some_num: int
    def inner():
        nonlocal some_num
        some_num = 5
    inner()
    print(some_num)


def type_annotation_only_gets_value_via_nonlocal_nested():
    """Similar, with nesting"""
    some_num: int
    def inner():
        def inner2():
            nonlocal some_num
            some_num = 5
        inner2()
    inner()
    print(some_num)


def type_annotation_never_gets_value_despite_nonlocal():
    """Type annotation lacks a value despite nonlocal declaration"""
    some_num: int
    def inner():
        nonlocal some_num
    inner()
    print(some_num)  # [used-before-assignment]


def inner_function_lacks_access_to_outer_args(args):
    """Check homonym between inner function and outer function names"""
    def inner():
        print(args)  # [used-before-assignment]
        args = []
    inner()
    print(args)


def inner_function_ok(args):
    """Explicitly redefined homonym defined before is OK."""
    def inner():
        args = []
        print(args)
    inner()
    print(args)


def nonlocal_in_outer_frame_fail():
    """Nonlocal declared in outer frame, bad usage and assignment in inner frame."""
    num = 1
    def outer():
        nonlocal num
        def inner():
            print(num)  # [used-before-assignment]
            num = 2
        inner()
    outer()


def nonlocal_in_outer_frame_ok(callback, condition_a, condition_b):
    """Nonlocal declared in outer frame, usage and definition in different frames,
    both enclosed in outer frame.
    """
    def outer():
        nonlocal callback
        if condition_a:
            def inner():
                callback()  # should not emit possibly-used-before-assignment
            inner()
        else:
            if condition_b:
                def callback():
                    pass
    outer()


def nonlocal_in_distant_outer_frame_fail(callback, condition_a, condition_b):
    """Nonlocal declared in outer frame, both usage and definition immediately enclosed
    in intermediate frame.
    """
    def outer():
        nonlocal callback
        def intermediate():
            if condition_a:
                def inner():
                    callback()  # [possibly-used-before-assignment]
                inner()
            else:
                if condition_b:
                    def callback():
                        pass
        intermediate()
    outer()


def nonlocal_after_bad_usage_fail():
    """Nonlocal declared after used-before-assignment."""
    num = 1
    def inner():
        num = num + 1  # [used-before-assignment]
        nonlocal num
    inner()



"""Tests for used-before-assignment when postponed evaluation of annotations is enabled"""
# pylint: disable=missing-function-docstring, invalid-name, missing-class-docstring, too-few-public-methods
from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    var = 1
    import math

print(var)  # [used-before-assignment]

def function_one(m: math):  # no error for annotations
    return m

# https://github.com/pylint-dev/pylint/issues/8893
if TYPE_CHECKING:
    import datetime

def f():
    return datetime.datetime.now()  # [used-before-assignment]

def g() -> datetime.datetime:
    return datetime.datetime.now()  # [used-before-assignment]

if TYPE_CHECKING:
    class X:
        pass

def h():
    return X()  # [used-before-assignment]

def i() -> X:
    return X()  # [used-before-assignment]

if TYPE_CHECKING:
    from mod import Y

def j():
    return {Y() for _ in range(1)}  # FALSE NEGATIVE



class Book:
    __match_args__ = ("title", "year")

    def __init__(self, title, year):
        self.title = title
        self.year = year


def func(item: Book):
    match item:
        case Book("abc", title="abc"):  # [multiple-class-sub-patterns]
            ...
        case Book(year=2000, year=2001):  # [multiple-class-sub-patterns]
            ...



# Licensed under the GPL: https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
# For details: https://github.com/pylint-dev/pylint/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/pylint/blob/main/CONTRIBUTORS.txt

"""Script used to generate the pyreverse configuration page."""

from __future__ import annotations

from pathlib import Path
from typing import NamedTuple

from sphinx.application import Sphinx

from pylint.pyreverse.main import OPTIONS_GROUPS, Run
from pylint.typing import OptionDict
from pylint.utils import get_rst_title


class OptionsData(NamedTuple):
    name: str
    optdict: OptionDict


PYREVERSE_PATH = (
    Path(__file__).resolve().parent.parent / "additional_tools" / "pyreverse"
)
"""Path to the pyreverse documentation folder."""


def _write_config_page(run: Run) -> None:
    """Create or overwrite the configuration page."""
    sections: list[str] = [
        f"""\
.. This file is auto-generated. Make any changes to the associated
.. docs extension in 'doc/exts/pyreverse_configuration.py'.


{get_rst_title("Usage", "#")}

``pyreverse`` is run from the command line using the following syntax::

  pyreverse [options] <packages>

where ``<packages>`` is one or more Python packages or modules to analyze.

The available options are organized into the following categories:

* :ref:`filtering-and-scope` - Control which classes and relationships appear in your diagrams
* :ref:`display-options` - Customize the visual appearance including colors and labels
* :ref:`output-control` - Select output formats and set the destination directory
* :ref:`project-configuration` - Define project settings like source roots and ignored files
"""
    ]
    options: list[OptionsData] = [OptionsData(name, info) for name, info in run.options]
    option_groups: dict[str, list[str]] = {g: [] for g in OPTIONS_GROUPS.values()}

    for option in sorted(options, key=lambda x: x.name):
        option_string = get_rst_title(f"--{option.name}", "-")
        option_string += f"*{option.optdict.get('help')}*\n\n"

        if option.optdict.get("default") == "":
            option_string += '**Default:** ``""``\n\n\n'
        else:
            option_string += f"**Default:**  ``{option.optdict.get('default')}``\n\n\n"

        option_groups[str(option.optdict.get("group"))].append(option_string)

    for group_title in OPTIONS_GROUPS.values():
        ref_title = group_title.lower().replace(" ", "-")
        sections.append(
            f"""\
.. _{ref_title}:

{get_rst_title(group_title, "=")}

{"".join(option_groups[group_title])}"""
        )

    # Join all sections and remove the final two newlines
    final_page = "\n\n".join(sections)[:-2]

    with open(PYREVERSE_PATH / "configuration.rst", "w", encoding="utf-8") as stream:
        stream.write(final_page)


# pylint: disable-next=unused-argument
def build_options_page(app: Sphinx | None) -> None:
    # Write configuration page
    _write_config_page(Run([]))


def setup(app: Sphinx) -> dict[str, bool]:
    """Connects the extension to the Sphinx process."""
    # Register callback at the builder-inited Sphinx event
    # See https://www.sphinx-doc.org/en/master/extdev/appapi.html
    app.connect("builder-inited", build_options_page)
    return {"parallel_read_safe": True}



# Test for https://github.com/pylint-dev/pylint/issues/9267
class A:
    def __init__(self) -> None:
        self.var = 2

class B:
    def __init__(self) -> None:
        self.a_obj = A()

    def func(self):
        self.a_obj = A()
        self.a_obj = A()



class P:
    def __init__(self, name: str):
        self.name = name


class PrivateAttr:
    def __init__(self):
        self.__x = P("private")


class ProtectedAttr:
    def __init__(self):
        self._x = P("protected")


class PublicAttr:
    def __init__(self):
        self.x = P("public")


class SpecialAttr:
    def __init__(self):
        self.__x__ = P("special")



# Licensed under the GPL: https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
# For details: https://github.com/pylint-dev/pylint/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/pylint/blob/main/CONTRIBUTORS.txt
# pylint: disable=missing-module-docstring, missing-function-docstring

from __future__ import annotations

import contextlib
import os
import pathlib
import shlex
import sys
from collections.abc import Sequence
from io import BufferedReader
from typing import Any, Protocol
from unittest.mock import MagicMock, mock_open, patch

import pytest

from pylint import run_pylint, run_pyreverse, run_symilar
from pylint.lint.run import _query_cpu
from pylint.testutils import GenericTestReporter as Reporter
from pylint.testutils._run import _Run as Run
from pylint.testutils.utils import _test_cwd


class _RunCallable(Protocol):  # pylint: disable=too-few-public-methods
    def __call__(self, argv: Sequence[str] | None = None) -> None: ...


@pytest.mark.parametrize("runner", [run_pylint, run_pyreverse, run_symilar])
def test_runner(runner: _RunCallable, tmp_path: pathlib.Path) -> None:
    filepath = os.path.abspath(__file__)
    testargs = ["", filepath]
    with _test_cwd(tmp_path):
        with patch.object(sys, "argv", testargs):
            with pytest.raises(SystemExit) as err:
                runner()
            assert err.value.code == 0


@pytest.mark.parametrize("runner", [run_pylint, run_pyreverse, run_symilar])
def test_runner_with_arguments(runner: _RunCallable, tmp_path: pathlib.Path) -> None:
    """Check the runners with arguments as parameter instead of sys.argv."""
    filepath = os.path.abspath(__file__)
    testargs = [filepath]
    with _test_cwd(tmp_path):
        with pytest.raises(SystemExit) as err:
            runner(testargs)
        assert err.value.code == 0


def test_pylint_argument_deduplication(
    tmp_path: pathlib.Path, tests_directory: pathlib.Path
) -> None:
    """Check that the Pylint runner does not over-report on duplicate
    arguments.

    See https://github.com/pylint-dev/pylint/issues/6242 and
    https://github.com/pylint-dev/pylint/issues/4053
    """
    filepath = str(tests_directory / "functional/t/too/too_many_branches.py")
    testargs = shlex.split("--report n --score n --max-branches 13")
    testargs.extend([filepath] * 4)
    exit_stack = contextlib.ExitStack()
    exit_stack.enter_context(_test_cwd(tmp_path))
    exit_stack.enter_context(patch.object(sys, "argv", testargs))
    err = exit_stack.enter_context(pytest.raises(SystemExit))
    with exit_stack:
        run_pylint(testargs)
    assert err.value.code == 0


@pytest.mark.parametrize(
    "quota,shares,period",
    [
        # Shares path
        ("-1", "2", ""),
        ("-1", "1023", ""),
        ("-1", "1024", ""),
        # Periods path
        ("100", "", "200"),
        ("999", "", "1000"),
        ("1000", "", "1000"),
    ],
)
def test_pylint_run_dont_crash_with_cgroupv1(
    tmp_path: pathlib.Path,
    quota: str,
    shares: str,
    period: str,
) -> None:
    """Check that the pylint runner does not crash if `pylint.lint.run._query_cpu`
    determines only a fraction of a CPU core to be available.
    """
    filepath = os.path.abspath(__file__)
    testargs = [filepath, "--jobs=0"]

    with _test_cwd(tmp_path):
        with pytest.raises(SystemExit) as err:
            with patch(
                "builtins.open",
                mock_cgroup_fs(quota=quota, shares=shares, period=period),
            ):
                with patch("pylint.lint.run.Path", mock_cgroup_path(v2=False)):
                    Run(testargs, reporter=Reporter())
        assert err.value.code == 0


@pytest.mark.parametrize(
    "contents",
    [
        "1 2",
        "max 100000",
    ],
)
def test_pylint_run_dont_crash_with_cgroupv2(
    tmp_path: pathlib.Path,
    contents: str,
) -> None:
    """Check that the pylint runner does not crash if `pylint.lint.run._query_cpu`
    determines only a fraction of a CPU core to be available.
    """
    filepath = os.path.abspath(__file__)
    testargs = [filepath, "--jobs=0"]

    with _test_cwd(tmp_path):
        with pytest.raises(SystemExit) as err:
            with patch("builtins.open", mock_cgroup_fs(max_v2=contents)):
                with patch("pylint.lint.run.Path", mock_cgroup_path(v2=True)):
                    Run(testargs, reporter=Reporter())
        assert err.value.code == 0


@pytest.mark.parametrize(
    "contents,expected",
    [
        ("50000 100000", 1),
        ("100000 100000", 1),
        ("200000 100000", 2),
        ("299999 100000", 2),
        ("300000 100000", 3),
        # Unconstrained cgroup
        ("max 100000", None),
    ],
)
def test_query_cpu_cgroupv2(
    tmp_path: pathlib.Path,
    contents: str,
    expected: int,
) -> None:
    """Check that `pylint.lint.run._query_cpu` generates realistic values in cgroupsv2
    systems.
    """
    with _test_cwd(tmp_path):
        with patch("builtins.open", mock_cgroup_fs(max_v2=contents)):
            with patch("pylint.lint.run.Path", mock_cgroup_path(v2=True)):
                cpus = _query_cpu()
                assert cpus == expected


@pytest.mark.parametrize(
    "quota,shares,period,expected",
    [
        # Shares path
        ("-1", "2", "", 1),
        ("-1", "1023", "", 1),
        ("-1", "1024", "", 1),
        ("-1", "2048", "", 2),
        # Periods path
        ("100", "", "200", 1),
        ("999", "", "1000", 1),
        ("1000", "", "1000", 1),
        ("2000", "", "1000", 2),
    ],
)
def test_query_cpu_cgroupv1(
    tmp_path: pathlib.Path,
    quota: str,
    shares: str,
    period: str,
    expected: int,
) -> None:
    """Check that `pylint.lint.run._query_cpu` generates realistic values in cgroupsv1
    systems.
    """
    with _test_cwd(tmp_path):
        with patch(
            "builtins.open", mock_cgroup_fs(quota=quota, shares=shares, period=period)
        ):
            with patch("pylint.lint.run.Path", mock_cgroup_path(v2=False)):
                cpus = _query_cpu()
                assert cpus == expected


def mock_cgroup_path(v2: bool) -> Any:
    def _mock_path(*args: str, **kwargs: Any) -> pathlib.Path:
        if args[0] == "/sys/fs/cgroup/cpu/cpu.cfs_period_us":
            return MagicMock(is_file=lambda: not v2)
        if args[0] == "/sys/fs/cgroup/cpu/cpu.shares":
            return MagicMock(is_file=lambda: not v2)
        if args[0] == "/sys/fs/cgroup/cpu/cpu.cfs_quota_us":
            return MagicMock(is_file=lambda: not v2)
        if args[0] == "/sys/fs/cgroup/cpu.max":
            return MagicMock(is_file=lambda: v2)
        return pathlib.Path(*args, **kwargs)

    return _mock_path


def mock_cgroup_fs(
    quota: str = "", shares: str = "", period: str = "", max_v2: str = ""
) -> Any:
    builtin_open = open

    def _mock_open(*args: Any, **kwargs: Any) -> BufferedReader:
        if args[0] == "/sys/fs/cgroup/cpu/cpu.cfs_quota_us":
            return mock_open(read_data=quota)(*args, **kwargs)  # type: ignore[no-any-return]
        if args[0] == "/sys/fs/cgroup/cpu/cpu.shares":
            return mock_open(read_data=shares)(*args, **kwargs)  # type: ignore[no-any-return]
        if args[0] == "/sys/fs/cgroup/cpu/cpu.cfs_period_us":
            return mock_open(read_data=period)(*args, **kwargs)  # type: ignore[no-any-return]
        if args[0] == "/sys/fs/cgroup/cpu.max":
            return mock_open(read_data=max_v2)(*args, **kwargs)  # type: ignore[no-any-return]
        return builtin_open(*args, **kwargs)  # type: ignore[no-any-return]

    return _mock_open



class Book:
    __match_args__ = ("title", "year")

    def __init__(self, title, year):
        self.title = title
        self.year = year


def func(item: Book):
    match item:
        case Book(title=str(title)):  # [match-class-bind-self]
            ...
        case Book(year=int(year)):  # [match-class-bind-self]
            ...



# +1: [too-many-positional-arguments]
def calculate_drag_force(velocity, area, density, drag_coefficient):
    """Each argument is positional-or-keyword unless a `/` or `*` is present."""
    return 0.5 * drag_coefficient * density * area * velocity**2


drag_force = calculate_drag_force(30, 2.5, 1.225, 0.47)



# pylint: disable=missing-module-docstring

# Should not trigger undefined-variable (#10562)
item = (1, 2, 3)
for item in item:
    print(item)

# Should trigger undefined-variable
for iteree in iteree:  # [undefined-variable]
    print(iteree)



"""Test for deleted index."""

# pylint: disable=missing-function-docstring

def test_deleted_index(letters: list[str]) -> None:
    for index, letter in enumerate(letters): # [unused-variable]
        del index
        print(letters[index]) # [unnecessary-list-index-lookup]



# pylint: disable=missing-docstring,unused-variable,too-few-public-methods
# pylint: disable=match-class-positional-attributes

from typing import NamedTuple

# -- Check __match_args__ definitions --
class A:
    __match_args__ = ("x",)

class B(A):
    __match_args__ = ("x", "y")

class C:
    __match_args__ = ["x", "y"]  # [invalid-match-args-definition]

class D:
    __match_args__ = ("x", 1)  # [invalid-match-args-definition]

class E:
    def f(self):
        __match_args__ = ["x"]


class Result(NamedTuple):
    # inherits from tuple -> match self
    x: int
    y: int


def f1(x):
    """Check too many positional sub-patterns"""
    match x:
        case A(1): ...
        case A(1, 2): ...  # [too-many-positional-sub-patterns]
        case B(1, 2): ...
        case B(1, 2, 3): ...  # [too-many-positional-sub-patterns]
        case int(1): ...
        case int(1, 2): ...  # [too-many-positional-sub-patterns]
        case tuple(1): ...
        case tuple(1, 2): ...  # [too-many-positional-sub-patterns]
        case tuple((1, 2)): ...
        case Result(1, 2): ...

def f2(x):
    """Check multiple sub-patterns for attribute"""
    match x:
        case A(1, x=1): ...  # [multiple-class-sub-patterns]
        case A(1, y=1): ...
        case A(x=1, x=2, x=3): ...  # [multiple-class-sub-patterns]

        # with invalid __match_args__ we can't detect duplicates with positional patterns
        case D(1, x=1): ...

        # If class name is undefined, we can't get __match_args__
        case NotDefined(1, x=1): ...  # [undefined-variable]

def f3(x):
    """Check class pattern with name binding to self."""
    match x:
        case int(y): ... # [match-class-bind-self]
        case int() as y: ...
        case int(2 as y): ...
        case str(y): ...   # [match-class-bind-self]
        case str() as y: ...
        case str("Hello" as y): ...
        case tuple(y, 2): ...  # pylint: disable=too-many-positional-sub-patterns
        case tuple((y, 2)): ...

def f4(x):
    """Check for positional attributes if keywords could be used."""
    # pylint: enable=match-class-positional-attributes
    match x:
        case int(2): ...
        case bool(True): ...
        case A(1): ...  # [match-class-positional-attributes]
        case A(x=1): ...
        case B(1, 2): ...  # [match-class-positional-attributes]
        case B(x=1, y=2): ...
        case Result(1, 2): ...
        case Result(x=1, y=2): ...



# pylint: disable=missing-function-docstring, missing-module-docstring, invalid-name, import-outside-toplevel, deprecated-method
import codecs
import contextlib
import multiprocessing
import pathlib
import subprocess
import sys
import tarfile
import tempfile
import threading
import urllib
import zipfile
from pathlib import Path

if sys.version_info >= (3, 14):
    from concurrent.futures.process import ProcessPoolExecutor
    from concurrent.futures.thread import ThreadPoolExecutor
else:
    from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor


def test_pathlib_open():
    _ = pathlib.Path("foo").open(encoding="utf8")  # [consider-using-with]
    _ = Path("foo").open(encoding="utf8")  # [consider-using-with]
    path = Path("foo")
    _ = path.open(encoding="utf8")  # [consider-using-with]
    with Path("foo").open(encoding="utf8") as file:  # must not trigger
        _ = file.read()


def test_codecs_open():
    fh = codecs.open("test.txt", "utf8")  # [consider-using-with]
    fh.close()


def test_urlopen():
    _ = urllib.request.urlopen("http://www.python.org")  # [consider-using-with]


def test_temporary_file():
    _ = tempfile.TemporaryFile("r")  # ambiguous with NamedTemporaryFile


def test_named_temporary_file():
    _ = tempfile.NamedTemporaryFile("r")  # [consider-using-with]


def test_spooled_temporary_file():
    _ = tempfile.SpooledTemporaryFile("r")  # [consider-using-with]


def test_temporary_directory():
    _ = tempfile.TemporaryDirectory()  # [consider-using-with]


def test_zipfile():
    myzip = zipfile.ZipFile("spam.zip", "w")  # [consider-using-with]
    _ = myzip.open("eggs.txt")  # [consider-using-with]


def test_pyzipfile():
    myzip = zipfile.PyZipFile("spam.zip", "w")  # [consider-using-with]

    with zipfile.PyZipFile("spam.zip", "w"):  # must not trigger
        pass

    _ = myzip.open("eggs.txt")  # [consider-using-with]

    with myzip.open("eggs.txt"):  # must not trigger
        pass


def test_tarfile():
    tf = tarfile.open("/tmp/test.tar", "w")  # [consider-using-with]
    tf.close()

    with tarfile.open("/tmp/test.tar", "w"):  # must not trigger
        pass

    tf = tarfile.TarFile("/tmp/test2.tar", "w")  # [consider-using-with]
    tf.close()

    with tarfile.TarFile("/tmp/test2.tar", "w"):  # must not trigger
        pass


def test_lock_acquisition():
    lock = threading.Lock()
    lock.acquire()  # [consider-using-with]
    lock.release()

    with lock:  # must not trigger
        pass

    rlock = threading.RLock()
    rlock.acquire()  # [consider-using-with]
    rlock.release()

    with rlock:  # must not trigger
        pass

    sema = threading.Semaphore()
    sema.acquire()  # [consider-using-with]
    sema.release()

    with sema:  # must not trigger
        pass

    bounded_sema = threading.BoundedSemaphore()
    bounded_sema.acquire()  # [consider-using-with]
    bounded_sema.release()

    with bounded_sema:  # must not trigger
        pass


@contextlib.contextmanager
def test_lock_acquisition_in_context_manager1():
    """
    The message must not be triggered if the resource allocation is done inside a context manager.
    """
    lock = threading.Lock()
    lock.acquire()  # must not trigger
    yield
    lock.release()


class MyLockContext:
    """
    The message must not be triggered if the resource allocation is done inside a context manager.
    """

    def __init__(self):
        self.lock = threading.Lock()

    def __enter__(self):
        self.lock.acquire()  # must not trigger

    def __exit__(self, exc_type, exc_value, traceback):
        self.lock.release()


def test_multiprocessing():
    # the different Locks provided by multiprocessing would be candidates
    # for consider-using-with as well, but they lead to InferenceErrors.
    _ = multiprocessing.Pool()  # [consider-using-with]
    with multiprocessing.Pool():
        pass

    manager = multiprocessing.managers.BaseManager()
    manager.start()  # [consider-using-with]
    with multiprocessing.managers.BaseManager():
        pass

    manager = multiprocessing.managers.SyncManager()
    manager.start()  # [consider-using-with]
    with multiprocessing.managers.SyncManager():
        pass


def test_popen():
    _ = subprocess.Popen("sh")  # [consider-using-with]
    with subprocess.Popen("sh"):
        pass


def test_suppress_in_exit_stack():
    """Regression test for issue #4654 (false positive)"""
    with contextlib.ExitStack() as stack:
        _ = stack.enter_context(
            open("/sys/firmware/devicetree/base/hwid,location", "r", encoding="utf-8")
        )  # must not trigger


def test_futures():
    """
    Regression test for issue #4689.
    ThreadPoolExecutor and ProcessPoolExecutor were formerly part of the callables that raised
    the R1732 message if used outside a with block, but there are legitimate use cases where
    Executor instances are used e.g. as a persistent background worker pool throughout the program.
    """
    thread_executor = ThreadPoolExecutor()
    thread_executor.submit(print, 1)
    process_executor = ProcessPoolExecutor()
    process_executor.submit(print, 2)
    thread_executor.shutdown()
    process_executor.shutdown()


pool = multiprocessing.Pool()  # must not trigger, as it is used later on
with pool:
    pass


global_pool = multiprocessing.Pool()  # must not trigger, will be used in nested scope


def my_nested_function():
    with global_pool:
        pass


# this must also work for tuple unpacking
pool1, pool2 = (
    multiprocessing.Pool(),  # must not trigger
    multiprocessing.Pool(),  # must not trigger
)

with pool1:
    pass

with pool2:
    pass

unused_pool1, unused_pool2 = (
    multiprocessing.Pool(),  # [consider-using-with]
    multiprocessing.Pool(),  # [consider-using-with]
)

used_pool, unused_pool = (
    multiprocessing.Pool(),  # must not trigger
    multiprocessing.Pool(),  # [consider-using-with]
)
with used_pool:
    pass

unused_pool, used_pool = (
    multiprocessing.Pool(),  # [consider-using-with]
    multiprocessing.Pool(),  # must not trigger
)
with used_pool:
    pass


def test_subscript_assignment():
    """
    Regression test for issue https://github.com/pylint-dev/pylint/issues/4732.
    If a context manager is assigned to a list or dict, we are not able to
    tell if / how the context manager is used later on, as it is not assigned
    to a variable or attribute directly.
    In this case we can only emit the message directly.
    """
    job_list = [None, None]
    job_list[0] = subprocess.Popen("ls")  # [consider-using-with]
    job_dict = {}
    job_dict["myjob"] = subprocess.Popen("ls")  # [consider-using-with]



"""Test deprecated modules using '__import__' builtin method."""
# pylint: disable=invalid-name

imp = __import__("mypackage")  # [deprecated-module]

imp_meth = __import__("mypackage").meth()  # [deprecated-module]

lib = "mypackage"
infer_imp = __import__(lib)  # [deprecated-module]



"""The other functional file has a max-complexity so low that it doesn't cover the
 case when the code is not too complex."""


def not_too_complex():
    """McCabe rating: 1"""
    return True


def too_complex(condition):  # [too-complex]
    """McCabe rating: 2"""
    if condition is True:
        return True
    return False



"""Check invalid value returned by __getnewargs_ex__ """

# pylint: disable=too-few-public-methods,missing-docstring,import-error,use-dict-literal,unnecessary-lambda-assignment,use-dict-literal
import six

from missing import Missing


class FirstGoodGetNewArgsEx:
    """__getnewargs_ex__ returns <type 'tuple'>"""

    def __getnewargs_ex__(self):
        return ((1,), {"2": "2"})


class SecondGoodGetNewArgsEx:
    """__getnewargs_ex__ returns <type 'tuple'>"""

    def __getnewargs_ex__(self):
        return (tuple(), dict())


class GetNewArgsExMetaclass(type):
    def __getnewargs_ex__(cls):
        return ((1,), {"2": "2"})


@six.add_metaclass(GetNewArgsExMetaclass)
class ThirdGoodGetNewArgsEx:
    """GetNewArgsEx through the metaclass."""


class FourthGoodGetNewArgsEx:
    """Test that `args` and `kwargs` (`Name` nodes) are inferred as tuples.

    https://github.com/pylint-dev/pylint/issues/10208
    """
    def __init__(self, boo, far, *, hoo, haha):
        self._foo = boo
        self._bar = far
        self._hoo = hoo
        self._haha = haha

    def __getnewargs_ex__(self):
        args = (self._foo, self._bar)
        kwargs = {'hoo': self._hoo,
                  'haha': self._haha}
        return args, kwargs


class FirstBadGetNewArgsEx:
    """ __getnewargs_ex__ returns an integer """

    def __getnewargs_ex__(self):  # [invalid-getnewargs-ex-returned]
        return 1


class SecondBadGetNewArgsEx:
    """ __getnewargs_ex__ returns tuple with incorrect arg length"""

    def __getnewargs_ex__(self):  # [invalid-getnewargs-ex-returned]
        return (tuple(1), dict(x="y"), 1)


class ThirdBadGetNewArgsEx:
    """ __getnewargs_ex__ returns tuple with wrong type for first arg """

    def __getnewargs_ex__(self):  # [invalid-getnewargs-ex-returned]
        return (dict(x="y"), dict(x="y"))


class FourthBadGetNewArgsEx:
    """ __getnewargs_ex__ returns tuple with wrong type for second arg """

    def __getnewargs_ex__(self):  # [invalid-getnewargs-ex-returned]
        return ((1, ), (1, ))


class FifthBadGetNewArgsEx:
    """ __getnewargs_ex__ returns tuple with wrong type for both args """

    def __getnewargs_ex__(self):  # [invalid-getnewargs-ex-returned]
        return ({'x': 'y'}, (2,))


class SixthBadGetNewArgsEx:
    """ __getnewargs_ex__ returns node which does not have 'value' in AST """

    def __getnewargs_ex__(self):  # [invalid-getnewargs-ex-returned]
        return lambda: (1, 2)


class AmbigousGetNewArgsEx:
    """ Uninferable return value """

    __getnewargs_ex__ = lambda self: Missing


class AnotherAmbiguousGetNewArgsEx:
    """Potential uninferable return value"""

    def __getnewargs_ex__(self):
        return tuple(Missing)



# Test for https://github.com/pylint-dev/pylint/issues/9045

class P:
    pass

class Association:
    x: P  # just type hint, no ownership → Association

class Aggregation1:
    def __init__(self, x: P):
        self.x = x  # receives object, not created → Aggregation

class Aggregation2:
    x: P
    def __init__(self, x: P):
        self.x = x  # receives object, not created → Aggregation

class Composition1:
    x: P
    def __init__(self):
        self.x = P()  # creates object → Composition

class Composition2:
    def __init__(self):
        self.x = P()  # creates object → Composition



# Test for https://github.com/pylint-dev/pylint/issues/10236
from collections.abc import Generator
from dataclasses import dataclass


class Component:
    """A component class."""
    def __init__(self, name: str):
        self.name = name

class AssociationContainer:
    """Type hints only - no ownership."""
    def __init__(self):
        # Association: just type hints, no actual assignment
        self.components: list[Component]
        self.component_dict: dict[int, Component]
        self.components_set: set[Component]
        self.lazy_components: Generator[Component]

class AggregationContainer:
    """Comprehensions using existing objects - aggregation."""
    def __init__(self, existing_components: list[Component]):
        # Aggregation: comprehensions using existing objects (not creating)
        self.components: list[Component] = [comp for comp in existing_components]
        self.component_dict: dict[str, Component] = {f"key_{i}": comp for i, comp in enumerate(existing_components)}
        self.components_set: set[Component] = {comp for comp in existing_components}
        self.lazy_components: Generator[Component] = (comp for comp in existing_components)

class CompositionContainer:
    """Comprehensions creating new objects - composition."""
    def __init__(self):
        # Composition: comprehensions creating new objects
        self.components: list[Component] = [Component(f"component_{i}") for i in range(3)]
        self.component_dict: dict[int, Component] = {i: Component(f"dict_component_{i}") for i in range(2)}
        self.components_set: set[Component] = {Component(f"set_component_{i}") for i in range(2)}
        self.lazy_components: Generator[Component] = (Component(f"lazy_{i}") for i in range(2))



import asyncio
from contextlib import asynccontextmanager


@asynccontextmanager
async def async_context():
    yield


async def main():
    async with async_context():
        print("This works correctly")



# property_decorator.py
class AnnotatedPropertyTest:
    """Test class for property decorators with annotated return type"""
    def __init__(self):
        self._x = 0

    @property
    def x(self) -> int:
        """This is a getter for x"""
        return self._x

    @x.setter
    def x(self, value):
        """This is a setter for x"""
        self._x = value

    @x.deleter
    def x(self):
        """This is a deleter for x"""
        del self._x


class NonAnnotatedPropertyTest:
    """Test class for property decorators without annotated return type"""
    def __init__(self):
        self._x = 0

    @property
    def x(self):
        """This is a getter for x"""
        return self._x

    @x.setter
    def x(self, value):
        """This is a setter for x"""
        self._x = value

    @x.deleter
    def x(self):
        """This is a deleter for x"""
        del self._x



"""Functional test for consider-math-not-float."""

inf_float = float("inf")  # [consider-math-not-float]
neg_inf_float = float('-inf')  # [consider-math-not-float]
pos_inf_float = float("+inf")  # [consider-math-not-float]
infinity_float = float("infinity")  # [consider-math-not-float]
neg_infinity_float = float("-infinity")  # [consider-math-not-float]
large_exp_float = float("1e1000")
neg_large_exp_float = float("-1e1000")
very_large_exp_float = float("2.5E9999")
invalid_inf_float = float("in")  # [consider-math-not-float]
invalid_float_call = float("in", base=10)  # [consider-math-not-float]
nan_float = float("nan")  # [consider-math-not-float]
neg_nan_float = float("-nan")  # [consider-math-not-float]
pos_nan_float = float("+nan")  # [consider-math-not-float]
upper_nan_float = float("NaN")  # [consider-math-not-float]
typo_nan_float = float("nani")  # [consider-math-not-float]
other_typo_nan_float = float("nna")  # [consider-math-not-float]



red = 0
green = 1
blue = 2


def func(color):
    match color:
        case red:  # [bare-name-capture-pattern]
            print("I see red!")
        case green:  # [bare-name-capture-pattern]
            print("Grass is green")
        case blue:
            print("I'm feeling the blues :(")



"""Check trailing comma tuple optimization."""
# pylint: disable=missing-docstring

aaa = 1,
bbb = "aaaa",
ccc="aaa",
fff=['f'],

def some_func(first, second):
    if first:
        return first,
    if second:
        return (first, second,)
    return first, second,

#pylint:enable = trailing-comma-tuple
aaa = 1,  # [trailing-comma-tuple]
bbb = "aaaa", # [trailing-comma-tuple]
# pylint: disable=trailing-comma-tuple
ccc="aaa",
III = some_func(0,
    0),
# pylint: enable=trailing-comma-tuple
fff=['f'], # [trailing-comma-tuple]



# pylint: disable=undefined-variable, use-list-literal, unnecessary-lambda-assignment, use-dict-literal, disallowed-name
"""test suspicious lambda expressions
"""


# Some simple examples of the most commonly encountered forms.
# +1: [unnecessary-lambda]
_ = lambda: list()  # replaceable with "list"
# +1: [unnecessary-lambda]
_ = lambda x: hash(x)  # replaceable with "hash"
# +1: [unnecessary-lambda]
_ = lambda x, y: min(x, y)  # replaceable with "min"

# A function that can take any arguments given to it.
_anyargs = lambda *args, **kwargs: 'completely arbitrary return value'

# Some more complex forms of unnecessary lambda expressions.
# +1: [unnecessary-lambda]
_ = lambda *args: _anyargs(*args)
# +1: [unnecessary-lambda]
_ = lambda **kwargs: _anyargs(**kwargs)
# +1: [unnecessary-lambda]
_ = lambda *args, **kwargs: _anyargs(*args, **kwargs)
# +1: [unnecessary-lambda]
_ = lambda x, y, z, *args, **kwargs: _anyargs(x, y, z, *args, **kwargs)

# These don't use their parameters in their body
# +1: [unnecessary-lambda]
_ = lambda x: z(lambda x: x)(x)
# +1: [unnecessary-lambda]
_ = lambda x, y: z(lambda x, y: x + y)(x, y)

# Lambdas that are *not* unnecessary and should *not* trigger warnings.
_ = lambda x: x
_ = lambda x: x()
_ = lambda x=4: hash(x)
_ = lambda x, y: list(range(y, x))
_ = lambda x: list(range(5, x))
_ = lambda x, y: list(range(x, 5))
_ = lambda x, y, z: x.y(z)
_ = lambda: 5
_ = lambda **kwargs: _anyargs()
_ = lambda **kwargs: _anyargs(**dict([('three', 3)]))
_ = lambda **kwargs: _anyargs(**{'three': 3})
_ = lambda dict_arg, **kwargs: _anyargs(kwargs, **dict_arg)
_ = lambda *args: _anyargs()
_ = lambda *args: _anyargs(*list([3, 4]))
_ = lambda *args: _anyargs(*[3, 4])
_ = lambda list_arg, *args: _anyargs(args, *list_arg)
_ = lambda: _anyargs(*[3])
_ = lambda: _anyargs(**{'three': 3})
_ = lambda: _anyargs(*[3], **{'three': 3})
_ = lambda: _anyargs(func=42)

# pylint: disable=missing-function-docstring
def f(d):
    print(lambda x: str(x, **d))

# Don't warn about this.
_ = lambda: code().analysis()

_ = lambda **kwargs: dict(bar=42, **kwargs)

# These use the lambda parameters in their body
_ = lambda x: x(x)
_ = lambda x, y: x(x, y)
_ = lambda x: z(lambda y: x + y)(x)


# https://github.com/pylint-dev/pylint/issues/8192

# foo does not yet exist, so replacing lambda x: foo.get(x) with
# foo.get will raise NameError
g = lambda x: foo.get(x)  # [unnecessary-lambda]  FALSE POSITIVE

# an object is created and given the name 'foo'
foo = {1: 2}
assert g(1) == 2

# a new object is created and given the name 'foo'; first object is lost
foo = {1: 3}
assert g(1) == 3

# the name 'foo' is deleted; second object is lost; there is no foo
del foo

assert g(1) == 3  # NameError: name 'foo' is not defined



from enum import Enum


class Color(Enum):
    RED = 0
    GREEN = 1
    BLUE = 2


def func(color: Color) -> None:
    match color:
        case Color.RED:
            print("I see red!")
        case Color.GREEN:
            print("Grass is green")
        case Color.BLUE:
            print("I'm feeling the blues :(")



# pylint: disable=missing-function-docstring, missing-module-docstring

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from datetime import datetime


def func_two():
    second = datetime.now()  # [used-before-assignment]
    return second


def func():
    first: datetime
    first = datetime.now()  # [used-before-assignment]
    second = datetime.now()
    return first, second


def annotated_declaration_with_type_subscript():
    dt_list: list[datetime]
    dt_list = [datetime.now(), datetime.now()]  # [used-before-assignment]

    return dt_list



# pylint: disable=invalid-name,unnecessary-pass,no-else-return,useless-else-on-loop
# pylint: disable=undefined-variable,consider-using-sys-exit,unused-variable,too-many-return-statements
# pylint: disable=redefined-outer-name,using-constant-test,unused-argument,unnecessary-lambda-assignment
# pylint: disable=broad-except, not-context-manager, no-method-argument, unspecified-encoding, broad-exception-raised

"""Checks use of "too-complex" check"""


def just_a_pass():  # [too-complex]
    """McCabe rating: 1"""
    pass


def just_a_yield():  # [too-complex]
    """McCabe rating: 1"""
    yield from range(10)


def just_a_return():  # [too-complex]
    """McCabe rating: 1"""
    return 42


def just_a_raise():  # [too-complex]
    """McCabe rating: 1"""
    raise ValueError("An error occurred")


def one_edge_multiple_operations(n):  # [too-complex]
    """McCabe rating: 1"""
    k = n + 4
    s = k + n
    return s


def if_elif_else(n):  # [too-complex]
    """McCabe rating: 3"""
    if n > 3:
        return "bigger than three"
    elif n > 4:
        return "is never executed"
    else:
        return "smaller than or equal to three"


def if_with_conditionals(a, b, c):  # [too-complex]
    """McCabe rating: 2"""
    if (  # pylint: disable=too-many-boolean-expressions
        a
        and b
        or c
        or (a and not b)
        or (b and not c)
        or (c and not a)
        or (a and b and c)
    ):
        return True
    return False


def for_loop():  # [too-complex]
    """McCabe rating: 2"""
    for i in range(10):
        print(i)


def for_loop_with_else(mylist):  # [too-complex]
    """McCabe rating: 2"""
    for i in mylist:
        print(i)
    else:
        print(None)


def recursive_if_else(n):  # [too-complex]
    """McCabe rating: 2"""
    if n > 4:
        return recursive_if_else(n - 1)
    else:
        return n


def for_loop_with_break():  # [too-complex]
    """McCabe rating: 3"""
    for i in range(10):
        if i == 5:
            break


def for_loop_with_continue():  # [too-complex]
    """McCabe rating: 3"""
    for i in range(10):
        if i % 2 == 0:
            continue
        print(i)


def for_loop_with_continue_and_break():  # [too-complex]
    """McCabe rating: 4"""
    for i in range(10):
        if i % 2 == 0:
            continue
        if i % 5 == 0:
            break


def inner_functions():  # [too-complex]
    """McCabe rating: 3"""

    def inner_function():  # Known false negative ?
        """McCabe rating: 2"""

        def innermost_function():  # Known false negative ?
            """McCabe rating: 1"""
            pass

        innermost_function()

    inner_function()


def try_with_multiple_except_and_else():  # [too-complex]
    """McCabe rating: 4"""
    try:
        print(1)
    except TypeA:
        print(2)
    except TypeB:
        print(3)
    else:
        print(4)


def with_(fp):  # [too-complex]
    """McCabe rating: 1"""
    with open(fp) as f:
        content = f.read()
    return content


def lambda_with_if(lst):  # [too-complex]
    """McCabe rating should be 4, but is 1 (known false negative ?)

    See counterpart 'comprehension_with_if' below."""
    f = lambda x: [x for x in lst if x % 2 == 0] or range(10)
    return f(lst)


def comprehension_with_if(lst):  # [too-complex]
    """McCabe rating: should be 4 but is 1 (known false negative ?)
    https://github.com/PyCQA/mccabe/issues/69
    """
    return [x for x in lst if x % 2 == 0] or range(10)


def comprehension_with_if_equivalent(lst):  # [too-complex]
    """McCabe rating: 4

    See counterpart 'comprehension_with_if' above.
    """
    xs = []
    for x in lst:
        if x % 2 == 0:
            xs.append(x)
    if xs:
        return xs
    else:
        return range(10)


def nested_ifs_elifs_elses():  # [too-complex]
    """McCabe rating: 9"""
    myint = 2
    if myint > 5:
        pass
    else:
        if myint <= 5:
            pass
        else:
            myint = 3
            if myint > 2:
                if myint > 3:
                    pass
                elif myint == 3:
                    pass
                elif myint < 3:
                    pass
                else:
                    if myint:
                        pass
            else:
                if myint:
                    pass
                myint = 4


def big_elif_chain_with_nested_ifs():  # [too-complex]
    """McCabe rating: 11"""
    myint = 2
    if myint == 5:
        return myint
    elif myint == 6:
        return myint
    elif myint == 7:
        return myint
    elif myint == 8:
        return myint
    elif myint == 9:
        return myint
    elif myint == 10:
        if myint == 8:
            while True:
                return True
        elif myint == 8:
            with myint:
                return 8
    else:
        if myint == 2:
            return myint
        return myint
    return myint


class ExampleComplexityClass:
    """Class of example to test mccabe"""

    _name = "ExampleComplexityKlass"  # To force a tail.node=None

    def just_a_pass(self):  # [too-complex]
        """McCabe rating: 1"""
        pass

    def highly_complex(self, param1):  # [too-complex, too-many-branches]
        """McCabe rating: 15"""
        if not param1:
            pass
        pass
        if param1:
            pass
        else:
            pass
        pass
        if param1:
            pass
        if param1:
            pass
        if param1:
            pass
        if param1:
            pass
        if param1:
            pass
        if param1:
            pass
        if param1:
            for value in range(5):
                pass
        pass
        for count in range(6):
            with open("myfile") as fp:
                count += 1
            pass
        pass
        try:
            pass
            if not param1:
                pass
            else:
                pass
            if param1:
                raise BaseException("Error")
            with open("myfile2") as fp2:
                pass
            pass
        finally:
            if param1 is not None:
                pass
            for count2 in range(8):
                try:
                    pass
                except BaseException("Error2"):
                    pass
        return param1


for count in range(10):  # [too-complex]
    # McCabe rating: 4
    if count == 1:
        exit(0)
    elif count == 2:
        exit(1)
    else:
        exit(2)


def try_finally_with_nested_ifs():  # [too-complex]
    """McCabe rating: 3"""
    try:
        if True:
            pass
        else:
            pass
    finally:
        pass
    return True


def match_case(avg):  # [too-complex]
    """McCabe rating: 4
    See https://github.com/astral-sh/ruff/issues/11421
    """
    # pylint: disable=bare-name-capture-pattern
    match avg:
        case avg if avg < 0.3:
            avg_grade = "F"
        case avg if avg < 0.7:
            avg_grade = "E+"
        case _:
            raise ValueError(f"Unexpected average: {avg}")
    return avg_grade


def nested_match_case(data):  # [too-complex]
    """McCabe rating: 8

    Nested match statements."""
    match data:
        case {"type": "user", "data": user_data}:
            match user_data:  # Nested match adds complexity
                case {"name": str() as name}:
                    return f"User: {name}"
                case {"id": int() as user_id}:
                    return f"User ID: {user_id}"
                case _:
                    return "Unknown user format"
        case {"type": "product", "data": product_data}:
            if "price" in product_data:  # +1 for if
                return f"Product costs {product_data['price']}"
            else:
                return "Product with no price"
        case _:
            return "Unknown data type"


def yield_in_for_loop(a=None, b=None, c=None):  # [too-complex]
    """McCabe rating: 4"""
    yield from a or ()
    for elt in b:
        if elt is not None:
            yield elt
    if c is not None:
        yield c



"""Test `unused-variable` is not emitted for either case of `__future__.annotations` or `__all__`"""


from __future__ import annotations


__all__ = [ "apple" ]


def apple():
    """A public function"""



"""Tests for used-before-assignment with Python 3.13 type var defaults (PEP 696)"""
# pylint: disable=missing-docstring,unused-argument,too-few-public-methods

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    class Y: ...

class Good1[T = Y]: ...
class Good2[*Ts = tuple[int, Y]]: ...
class Good3[**P = [int, Y]]: ...
type Alias[T = Y] = T | None

# https://github.com/pylint-dev/pylint/issues/9884
def func[T = Y](x: T) -> None:  # [redefined-outer-name]  FALSE POSITIVE
    ...



# Licensed under the GPL: https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
# For details: https://github.com/pylint-dev/pylint/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/pylint/blob/main/CONTRIBUTORS.txt

"""Tests for the functional test framework."""

import contextlib
import os
import os.path
import shutil
import tempfile
from collections.abc import Iterator
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest
from _pytest.outcomes import Skipped

from pylint import testutils
from pylint.testutils.functional import (
    FunctionalTestFile,
    get_functional_test_files_from_directory,
)

HERE = Path(__file__).parent
DATA_DIRECTORY = HERE / "data"


@contextlib.contextmanager
def tempdir() -> Iterator[str]:
    """Create a temp directory and change the current location to it.

    This is supposed to be used with a *with* statement.
    """
    tmp = tempfile.mkdtemp()

    # Get real path of tempfile, otherwise test fail on mac os x
    current_dir = os.getcwd()
    os.chdir(tmp)
    abs_tmp = os.path.abspath(".")

    try:
        yield abs_tmp
    finally:
        os.chdir(current_dir)
        shutil.rmtree(abs_tmp)


@pytest.fixture(name="pytest_config")
def pytest_config_fixture() -> MagicMock:
    def _mock_getoption(option: str) -> bool:
        if option == "minimal_messages_config":
            return True
        return False

    config = MagicMock()
    config.getoption.side_effect = _mock_getoption
    return config


def test_parsing_of_pylintrc_init_hook() -> None:
    """Test that we correctly parse an init-hook in a settings file."""
    with pytest.raises(RuntimeError):
        test_file = FunctionalTestFile(str(DATA_DIRECTORY), "init_hook.py")
        testutils.LintModuleTest(test_file)


def test_get_functional_test_files_from_directory() -> None:
    """Test that we correctly check the functional test directory structures."""
    with pytest.raises(AssertionError) as exc_info:
        get_functional_test_files_from_directory(DATA_DIRECTORY / "u")
    assert exc_info.match("'use_dir.py' should go in 'use'")
    assert exc_info.match(
        "using_dir.py should go in a directory that starts with the "
        "first letters of 'using_dir'"
    )
    assert "incredibly_bold_mischief.py" not in str(exc_info.value)
    # Leading underscore mean that this should not fail the assertion
    get_functional_test_files_from_directory(DATA_DIRECTORY / "u/_no_issue_here")


def test_get_functional_test_files_from_crowded_directory() -> None:
    """Test that we correctly check the functional test directory structures."""
    with pytest.raises(AssertionError) as exc_info:
        get_functional_test_files_from_directory(
            DATA_DIRECTORY / "m", max_file_per_directory=1
        )
    assert exc_info.match("m: 3 when the max is 1")
    assert exc_info.match("max_overflow: 2 when the max is 1")
    assert len(exc_info.value.args[0].splitlines()) == 3
    with pytest.raises(AssertionError) as exc_info:
        get_functional_test_files_from_directory(
            DATA_DIRECTORY / "m", max_file_per_directory=2
        )
    assert exc_info.match("m: 3 when the max is 2")
    assert "max_overflow" not in str(exc_info.value)


@pytest.mark.parametrize(
    ["files", "output_file_name"],
    [
        ([], "file.txt"),
        (["file.txt"], "file.txt"),
        (["file.314.txt"], "file.txt"),  # don't match 3.14
        (["file.42.txt"], "file.txt"),  # don't match 4.2
        (["file.32.txt", "file.txt"], "file.32.txt"),
        (["file.312.txt", "file.txt"], "file.312.txt"),
        (["file.313.txt", "file.txt"], "file.313.txt"),
        (["file.310.txt", "file.313.txt", "file.312.txt", "file.txt"], "file.313.txt"),
        # don't match other test file names accidentally
        ([".file.313.txt"], "file.txt"),
        (["file_other.313.txt"], "file.txt"),
        (["other_file.313.txt"], "file.txt"),
    ],
)
def test_expected_output_file_matching(files: list[str], output_file_name: str) -> None:
    """Test output file matching. Pin current Python version to 3.13."""
    with tempdir():
        for file in files:
            with open(file, "w", encoding="utf-8"):
                ...
        test_file = FunctionalTestFile(".", "file.py")
        with patch(
            "pylint.testutils.functional.test_file._CURRENT_VERSION",
            new=(3, 13),
        ):
            assert test_file.expected_output == f".{os.path.sep}{output_file_name}"


def test_minimal_messages_config_enabled(pytest_config: MagicMock) -> None:
    """Test that all messages not targeted in the functional test are disabled
    when running with --minimal-messages-config.
    """
    test_file = FunctionalTestFile(
        str(DATA_DIRECTORY / "m"), "minimal_messages_config.py"
    )
    mod_test = testutils.LintModuleTest(test_file, pytest_config)
    assert all(
        mod_test._linter.is_message_enabled(msgid)
        for msgid in (
            "consider-using-with",
            "unspecified-encoding",
            "consider-using-f-string",
            # Always enable fatal errors: important not to have false negatives
            "astroid-error",
            "fatal",
            "syntax-error",
        )
    )
    assert not mod_test._linter.is_message_enabled("unused-import")


def test_minimal_messages_config_excluded_file(pytest_config: MagicMock) -> None:
    """Test that functional test files can be excluded from the run with
    --minimal-messages-config if they set the exclude_from_minimal_messages_config
    option in their rcfile.
    """
    test_file = FunctionalTestFile(
        str(DATA_DIRECTORY / "m"), "minimal_messages_excluded.py"
    )
    mod_test = testutils.LintModuleTest(test_file, pytest_config)
    with pytest.raises(Skipped):
        mod_test.setUp()



"""Tests for the useless-with-lock message"""
# pylint: disable=missing-docstring
import threading
from threading import Lock, RLock, Condition, Semaphore, BoundedSemaphore


with threading.Lock():  # [useless-with-lock]
    ...

with Lock():  # [useless-with-lock]
    ...

with threading.Lock() as THIS_SHOULDNT_MATTER:  # [useless-with-lock]
    ...

with threading.RLock():  # [useless-with-lock]
    ...

with RLock():  # [useless-with-lock]
    ...

with threading.Condition():  # [useless-with-lock]
    ...

with Condition():  # [useless-with-lock]
    ...

with threading.Semaphore():  # [useless-with-lock]
    ...

with Semaphore():  # [useless-with-lock]
    ...

with threading.BoundedSemaphore():  # [useless-with-lock]
    ...

with BoundedSemaphore():  # [useless-with-lock]
    ...

LOCK = threading.Lock()
with LOCK:  # this is ok
    ...

RLOCK = threading.RLock()
with RLOCK:  # this is ok
    ...

COND = threading.Condition()
with COND:  # this is ok
    ...

SEM = threading.Semaphore()
with SEM:  # this is ok
    ...

B_SEM = threading.BoundedSemaphore()
with B_SEM:  # this is ok
    ...



# pylint: disable=missing-docstring, invalid-name, unused-argument, consider-using-sys-exit, no-else-return
"""Test that quit() and exit() are handled consistently with sys.exit()"""

import sys

# These functions should not trigger inconsistent-return-statements
# because quit, exit, and sys.exit are never-returning functions (#10508)

def func_with_quit_1(i):
    """quit() in else branch should be treated like sys.exit()"""
    if i == 1:
        return i
    quit(1)

def func_with_quit_2(i):
    """quit() in if branch should be treated like sys.exit()"""
    if i == 1:
        quit(1)
    return 1

def func_with_exit_1(i):
    """exit() in else branch should be treated like sys.exit()"""
    if i == 1:
        return i
    exit(1)

def func_with_exit_2(i):
    """exit() in if branch should be treated like sys.exit()"""
    if i == 1:
        exit(1)
    return 1

def func_with_sys_exit_1(i):
    """sys.exit() in else branch - baseline test"""
    if i == 1:
        return i
    sys.exit(1)

def func_with_sys_exit_2(i):
    """sys.exit() in if branch - baseline test"""
    if i == 1:
        sys.exit(1)
    return 1

# Test mixed usage
def func_mixed_exit_methods(i):
    """Using different exit methods should all work consistently"""
    if i == 1:
        return "one"
    if i == 2:
        quit()
    if i == 3:
        exit()
    sys.exit()

# This should trigger inconsistent-return-statements
def func_inconsistent_example(i):  # [inconsistent-return-statements]
    """This should trigger the warning as a negative test case"""
    if i == 1:
        return i
    print("Not exiting, just printing")



# pylint: disable=missing-docstring,too-few-public-methods

class DynamicGetitem:
    def __getitem__(self, key):
        if key == 'attributes':
            return []
        return {'world': 123}


ex = DynamicGetitem()
a = ex['hello']['world']  # [invalid-sequence-index] known false-positive



"""Indirect tests that ExceptionGroup is inferred."""

try:
    ex_list = [ValueError("a"), ValueError("b"),  Exception("c")]
    raise ExceptionGroup("exceptions!", ex_list)
except* ValueError as exc:
    assert exc.exceptions
    assert exc.message



# pylint: disable=missing-module-docstring, missing-class-docstring, missing-function-docstring, too-few-public-methods

from functools import singledispatch, singledispatchmethod

# --- singledispatch function case ---
@singledispatch
def process(value):
    return f"default handler for {value!r}"

@process.register(int)
def _(value):
    return f"int handler for {value}"

@process.register(str)
def _(value):
    return f"str handler for {value}"

# --- singledispatchmethod case ---
class Handler:
    @singledispatchmethod
    def handle(self, value):
        return f"default method handler for {value!r}"

    @handle.register(int)
    def _(self, value):
        return f"int method handler for {value}"

    @handle.register(str)
    def _(self, value):
        return f"str method handler for {value}"



# pylint: disable=literal-comparison,missing-docstring

X = ''
Y = 'test'

if X is '':  # [use-implicit-booleaness-not-comparison-to-string]
    pass

if Y is not "":  # [use-implicit-booleaness-not-comparison-to-string]
    pass

if X == "":  # [use-implicit-booleaness-not-comparison-to-string]
    pass

if Y != '':  # [use-implicit-booleaness-not-comparison-to-string]
    pass

if "" == Y:  # [use-implicit-booleaness-not-comparison-to-string]
    pass

if '' != X:  # [use-implicit-booleaness-not-comparison-to-string]
    pass

if X == Y == "":
    pass

if "" == X == Y:
    pass

if X == Y == X == Y == "":
    pass


def test_in_boolean_context():
    """Cases where a comparison like `x != ""` is used in a boolean context.

    It is safe and idiomatic to simplify `x != ""` to just `x`.
    """
    # pylint: disable=pointless-statement,superfluous-parens,unnecessary-negation

    # Control flow
    if X != "":  # [use-implicit-booleaness-not-comparison-to-string]
        pass
    while X != "":  # [use-implicit-booleaness-not-comparison-to-string]
        pass
    assert X != ""  # [use-implicit-booleaness-not-comparison-to-string]

    # Ternary
    _ = 1 if X != "" else 2  # [use-implicit-booleaness-not-comparison-to-string]

    # Not
    if not (X != ""):  # [use-implicit-booleaness-not-comparison-to-string]
        pass

    # Comprehension filters
    [x for x in [] if X != ""]  # [use-implicit-booleaness-not-comparison-to-string]
    {x for x in [] if X != ""}  # [use-implicit-booleaness-not-comparison-to-string]
    (x for x in [] if X != "")  # [use-implicit-booleaness-not-comparison-to-string]

    # all() / any() with generator expressions
    all(X != "" for _ in range(1))  # [use-implicit-booleaness-not-comparison-to-string]
    any(X != "" for _ in range(1))  # [use-implicit-booleaness-not-comparison-to-string]

    # filter() with lambda
    filter(lambda: X != "", [])  # [use-implicit-booleaness-not-comparison-to-string]

    # boolean cast
    bool(X != "")  # [use-implicit-booleaness-not-comparison-to-string]

    # Logical operators nested in boolean contexts
    if X != "" and input():  # [use-implicit-booleaness-not-comparison-to-string]
        pass
    while input() or X != "":  # [use-implicit-booleaness-not-comparison-to-string]
        pass
    if (X != "" or input()) and input():  # [use-implicit-booleaness-not-comparison-to-string]
        pass


def test_not_in_boolean_context():
    """Cases where a comparison like `x != ""` is used in a non-boolean context.

    These comparisons cannot be safely replaced with just `x`, and should be explicitly
    cast using `bool(x)`.
    """
    # pylint: disable=pointless-statement
    _ = X != ""  # [use-implicit-booleaness-not-comparison-to-string]

    _ = X != "" or input()  # [use-implicit-booleaness-not-comparison-to-string]

    print(X != "")  # [use-implicit-booleaness-not-comparison-to-string]

    [X != "" for _ in []]  # [use-implicit-booleaness-not-comparison-to-string]

    lambda: X != ""  # [use-implicit-booleaness-not-comparison-to-string]

    filter(lambda x: x, [X != ""])  # [use-implicit-booleaness-not-comparison-to-string]

    return X != ""  # [use-implicit-booleaness-not-comparison-to-string]



class Book:
    __match_args__ = ("title", "year")

    def __init__(self, title, year, author):
        self.title = title
        self.year = year
        self.author = author


def func(item: Book):
    match item:
        case Book("title", 2000, author="author"):
            ...



# pylint: disable=literal-comparison,missing-docstring, singleton-comparison

X = 123
Y = len('test')

if X is 0:  # [use-implicit-booleaness-not-comparison-to-zero]
    pass

if X is False:
    pass

if Y is not 0:  # [use-implicit-booleaness-not-comparison-to-zero]
    pass

if Y is not False:
    pass

if X == 0:  # [use-implicit-booleaness-not-comparison-to-zero]
    pass

if X == False:
    pass

if 0 == Y:  # [use-implicit-booleaness-not-comparison-to-zero]
    pass

if Y != 0:  # [use-implicit-booleaness-not-comparison-to-zero]
    pass

if 0 != X:  # [use-implicit-booleaness-not-comparison-to-zero]
    pass

if Y != False:
    pass

if X > 0:
    pass

if X < 0:
    pass

if 0 < X:
    pass

if 0 > X:
    pass

if X == Y == 0:
    pass

if 0 == X == Y:
    pass

if X == Y == X == Y == 0:
    pass


def test_in_boolean_context():
    """Cases where a comparison like `x != 0` is used in a boolean context.

    It is safe and idiomatic to simplify `x != 0` to just `x`.
    """
    # pylint: disable=pointless-statement,superfluous-parens,unnecessary-negation

    # Control flow
    if X != 0:  # [use-implicit-booleaness-not-comparison-to-zero]
        pass
    while X != 0:  # [use-implicit-booleaness-not-comparison-to-zero]
        pass
    assert X != 0  # [use-implicit-booleaness-not-comparison-to-zero]

    # Ternary
    _ = 1 if X != 0 else 2  # [use-implicit-booleaness-not-comparison-to-zero]

    # Not
    if not (X != 0):  # [use-implicit-booleaness-not-comparison-to-zero]
        pass

    # Comprehension filters
    [x for x in [] if X != 0]  # [use-implicit-booleaness-not-comparison-to-zero]
    {x for x in [] if X != 0}  # [use-implicit-booleaness-not-comparison-to-zero]
    (x for x in [] if X != 0)  # [use-implicit-booleaness-not-comparison-to-zero]

    # all() / any() with generator expressions
    all(X != 0 for _ in range(1))  # [use-implicit-booleaness-not-comparison-to-zero]
    any(X != 0 for _ in range(1))  # [use-implicit-booleaness-not-comparison-to-zero]

    # filter() with lambda
    filter(lambda: X != 0, [])  # [use-implicit-booleaness-not-comparison-to-zero]

    # boolean cast
    bool(X != 0)  # [use-implicit-booleaness-not-comparison-to-zero]

    # Logical operators nested in boolean contexts
    if X != 0 and input():  # [use-implicit-booleaness-not-comparison-to-zero]
        pass
    while input() or X != 0:  # [use-implicit-booleaness-not-comparison-to-zero]
        pass
    if (X != 0 or input()) and input():  # [use-implicit-booleaness-not-comparison-to-zero]
        pass


def test_not_in_boolean_context():
    """Cases where a comparison like `x != 0` is used in a non-boolean context.

    These comparisons cannot be safely replaced with just `x`, and should be explicitly
    cast using `bool(x)`.
    """
    # pylint: disable=pointless-statement
    _ = X != 0  # [use-implicit-booleaness-not-comparison-to-zero]

    _ = X != 0 or input()  # [use-implicit-booleaness-not-comparison-to-zero]

    print(X != 0)  # [use-implicit-booleaness-not-comparison-to-zero]

    [X != 0 for _ in []]  # [use-implicit-booleaness-not-comparison-to-zero]

    lambda: X != 0  # [use-implicit-booleaness-not-comparison-to-zero]

    filter(lambda x: x, [X != 0])  # [use-implicit-booleaness-not-comparison-to-zero]

    return X != 0  # [use-implicit-booleaness-not-comparison-to-zero]



"""Functional tests for the ``bare-name-capture-pattern`` message"""


a = "a"
b = "b"
s = "a"


match s:
    case a:  # [bare-name-capture-pattern]
        pass
    case b:  # [bare-name-capture-pattern]
        pass
    case c if c == "Hello":
        pass
    case "a" as some_name:
        pass
    case s:
        pass



"""Tests for used-before-assignment with Python 3.12 generic typing syntax (PEP 695)"""
# pylint: disable = invalid-name,missing-docstring,too-few-public-methods,unused-argument

from typing import TYPE_CHECKING, Callable

type Point[T] = tuple[T, ...]
type Alias[*Ts] = tuple[*Ts]
type Alias2[**P] = Callable[P, None]

type AliasType = int | X | Y

class X:
    pass

if TYPE_CHECKING:
    class Y: ...

class Good[T: Y]: ...
type OtherAlias[T: Y] = T | None

# https://github.com/pylint-dev/pylint/issues/9884
def func[T: Y](x: T) -> None:  # [redefined-outer-name]  FALSE POSITIVE
    ...



# [line-too-long, missing-module-docstring, undefined-all-variable, undefined-all-variable, undefined-all-variable, undefined-all-variable, undefined-all-variable]
# ↓ equivalent to __all__ = ("C", "O", "N", "S", "T")
__all__ = tuple("CONST")

CONST = 42



# Licensed under the GPL: https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
# For details: https://github.com/pylint-dev/pylint/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/pylint/blob/main/CONTRIBUTORS.txt

from __future__ import annotations

import os
import sys
from datetime import datetime

# Pylint documentation build configuration file, created by
# sphinx-quickstart on Thu Apr  4 20:31:25 2013.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.


# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use 'os.path.abspath' to make it absolute, like shown here.
sys.path.append(os.path.abspath("exts"))
sys.path.append(os.path.abspath(".."))

# pylint: disable=wrong-import-position
from pylint import __version__
from pylint.__pkginfo__ import numversion

# pylint: enable=wrong-import-position

# -- General configuration -----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "pylint_features",
    "pylint_extensions",
    "pylint_messages",
    "pylint_options",
    "pyreverse_configuration",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.intersphinx",
    "sphinx_reredirects",
]


# Single file redirects are handled in this file and can be done by a pylint
# contributor. We use the following extension:
# https://documatt.gitlab.io/sphinx-reredirects/usage.html
# Directory redirects are handled in ReadTheDoc admin interface and can only be done
# by pylint maintainers at the following URL:
# https://readthedocs.org/dashboard/pylint/redirects/
redirects: dict[str, str] = {
    # "<source>": "<target>"
    "additional_commands/index": "../index.html",
    "development_guide/index": "api/index.html",
    "development_guide/contribute": "../development_guide/contributor_guide/index.html",
    "development_guide/contributor_guide": "contributor_guide/index.html",
    "development_guide/profiling": "../development_guide/contributor_guide/profiling.html",
    "development_guide/tests/index": "../contributor_guide/tests/index.html",
    "development_guide/tests/install": "../contributor_guide/tests/install.html",
    "development_guide/tests/launching_test": "../contributor_guide/tests/launching_test.html",
    # There was a typo in the original file, don't fix.
    "development_guide/tests/writting_test": "../contributor_guide/tests/writing_test.html",
    "development/testing": "tests/index.html",
    "how_tos/custom_checkers": "../development_guide/how_tos/custom_checkers.html",
    "how_tos/index": "../development_guide/how_tos/index.html",
    "how_tos/plugins": "../development_guide/how_tos/plugins.html",
    "how_tos/transform_plugins": "../development_guide/how_tos/transform_plugins.html",
    "intro": "index.html",
    "messages/messages_introduction": "../user_guide/messages/index.html",
    "messages/messages_list": "../user_guide/messages/messages_overview.html",
    "support": "contact.html",
    "technical_reference/c_extensions": "../user_guide/messages/error/no-member.html",
    "technical_reference/extensions": "../user_guide/checkers/extensions.html",
    "technical_reference/checkers": "../development_guide/technical_reference/checkers.html",
    "technical_reference/features": "../user_guide/checkers/features.html",
    "technical_reference/index": "../development_guide/technical_reference/index.html",
    "technical_reference/startup": "../development_guide/technical_reference/startup.html",
    "user_guide/configuration/naming-styles": "../user_guide/messages/convention/invalid-name.html",
    "user_guide/ide_integration/flymake-emacs": "../installation/ide_integration/flymake-emacs.html",
    "user_guide/ide_integration/ide-integration": "../installation/ide_integration/index.html",
    "user_guide/ide-integration": "installation.html",
    "user_guide/ide_integration/textmate": "../installation/ide_integration/textmate.html",
    "user_guide/index": "installation/index.html",
    "user_guide/message-control": "messages/message_control.html",
    "user_guide/options": "configuration/all-options.html",
    "user_guide/output": "usage/output.html",
    "user_guide/pre-commit-integration": "installation/pre-commit-integration.html",
    "user_guide/run": "usage/run.html",
    "pyreverse": "additional_tools/pyreverse/index.html",
    "symilar": "additional_tools/symilar/index.html",
}


# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The root toctree document.
root_doc = "index"

# General information about the project.
project = "Pylint"
current_year = datetime.utcnow().year
contributors = "Logilab and Pylint contributors"
copyright = f"2003-{current_year}, {contributors}"  # pylint: disable=redefined-builtin

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
# The short X.Y version.
version = f"{numversion[0]}.{numversion[1]}"
# The full version, including alpha/beta/rc tags.
release = __version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build", "data/**", "whatsnew/fragments"]

# The reST default role (used for this markup: `text`) to use for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []


# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "furo"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# Currently we use the default Furo configuration
# html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = "%b %d, %Y"

smartquotes = False

# Custom sidebar templates, maps document names to template names.
# Currently we use the default Furo Sidebar
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = "Pylintdoc"


# -- Options for LaTeX output --------------------------------------------------

# The paper size ('letter' or 'a4').
# latex_paper_size = 'letter'

# The font size ('10pt', '11pt' or '12pt').
# latex_font_size = '10pt'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
    (
        "index",
        "Pylint.tex",
        "Pylint Documentation",
        contributors,
        "manual",
    )
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Additional stuff for the LaTeX preamble.
# latex_preamble = ''

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [("index", "pylint", "Pylint Documentation", [contributors], 1)]

# pylint: disable-next=consider-using-namedtuple-or-dataclass
intersphinx_mapping = {
    "astroid": ("https://pylint.readthedocs.io/projects/astroid/en/latest/", None),
    "python": ("https://docs.python.org/3", None),
}

# Prevent label issues due to colliding section names
# through including multiple documents
autosectionlabel_prefix_document = True

# Permit duplicated titles in the resulting document.
# See https://github.com/pylint-dev/pylint/issues/7362#issuecomment-1256932866
autosectionlabel_maxdepth = 2

linkcheck_ignore = [
    "https://github.com/pylint-dev/pylint/blob/main/pylint/extensions/.*"
]



"""Check multiple key definition"""
# pylint: disable=pointless-statement, redundant-u-string-prefix

from enum import Enum


class MyEnum(Enum):
    """ Sample Enum for testing duplicate keys"""
    KEY = "key"



CORRECT_DICT = {
    'tea': 'for two',
    'two': 'for tea',
}

WRONG_WITH_ENUM = {  # [duplicate-key]
    MyEnum.KEY: "value 1",
    MyEnum.KEY: "value 2",
}

WRONG_DICT = {  # [duplicate-key]
    'tea': 'for two',
    'two': 'for tea',
    'tea': 'time',

}

{1: b'a', 1: u'a'} # [duplicate-key]
{1: 1, 1.0: 2} # [duplicate-key]



# pylint: disable=too-few-public-methods,missing-docstring,used-before-assignment,import-error,unused-import
# pylint: disable=wrong-import-position


@decorator
class DecoratedClass:
    decorator: int

import decorator



# pylint: disable=missing-function-docstring
"""Test async context manager used with regular 'with'."""

from contextlib import asynccontextmanager


@asynccontextmanager
async def async_cm():
    yield


with async_cm():  # [async-context-manager-with-regular-with]
    pass



"""Test deprecated modules from Python 3.9,
but use an earlier --py-version and ensure a warning is still emitted.
"""
# pylint: disable=unused-import, import-error

import binhex  # [deprecated-module]



"""PEP 695 generic typing nodes"""

from collections.abc import Callable, Sequence

type Point[T] = tuple[T, ...]
type Point[t] = tuple[t, ...]  # [invalid-name]

# Don't report typevar-name-incorrect-variance for type parameter
# The variance is determined by the type checker
type Array[T_co] = Sequence[T_co]

type Call[**_P] = Callable[_P, None]
type Call[**p] = Callable[p, None]  # [invalid-name]

type Tpl[*_Ts] = tuple[_Ts]
type Tpl[*ts] = tuple[ts]  # [invalid-name]



from contextlib import asynccontextmanager


@asynccontextmanager
async def async_context():
    yield


with async_context():  # [async-context-manager-with-regular-with]
    print("This will cause an error at runtime")



"""Test for old ternary constructs"""
from UNINFERABLE import condition, some_callable, maybe_true, maybe_false  # pylint: disable=import-error

TRUE_VALUE = True
FALSE_VALUE = False

SOME_VALUE1 = TRUE_VALUE if condition else FALSE_VALUE
SOME_VALUE2 = condition and TRUE_VALUE or FALSE_VALUE  # [consider-using-ternary]
NOT_SIMPLIFIABLE_1 = maybe_true if condition else maybe_false
NOT_SIMPLIFIABLE_2 = condition and maybe_true or maybe_false
SOME_VALUE3 = condition

def func1():
    """Ternary return value correct"""
    return TRUE_VALUE if condition else FALSE_VALUE


def func2():
    """Ternary return value incorrect"""
    return condition and TRUE_VALUE or FALSE_VALUE  # [consider-using-ternary]


SOME_VALUE4 = some_callable(condition) and 'ERROR' or 'SUCCESS'  # [consider-using-ternary]
SOME_VALUE5 = SOME_VALUE1 > 3 and 'greater' or 'not greater'  # [consider-using-ternary]
SOME_VALUE6 = (SOME_VALUE2 > 4 and SOME_VALUE3) and 'both' or 'not'  # [consider-using-ternary]
SOME_VALUE7 = 'both' if (SOME_VALUE2 > 4) and (SOME_VALUE3) else 'not'
SOME_VALUE8 = SOME_VALUE1 and SOME_VALUE2 and SOME_VALUE3 or SOME_VALUE4
SOME_VALUE9 = SOME_VALUE1 and False or SOME_VALUE2  # [simplify-boolean-expression]

YEAR = 1992
# Cannot be simplified with a ternary.
IS_LEAP_YEAR = YEAR % 4 == 0 and YEAR % 100 != 0 or YEAR % 400 == 0


def func4():
    """"Using a Name as a condition but still emits"""
    truth_value = 42
    return condition and truth_value or FALSE_VALUE # [consider-using-ternary]


def func5():
    """"Using a Name that infers to False as a condition does not emit"""
    falsy_value = False
    return condition and falsy_value or FALSE_VALUE # [simplify-boolean-expression]


def func_control_flow():
    """Redefining variables should invalidate simplify-boolean-expression."""
    flag_a = False
    flag_b = False
    for num in range(2):
        if num == 1:
            flag_a = True
        else:
            flag_b = True
    multiple = (flag_a and flag_b) or func5()
    return multiple



# Licensed under the GPL: https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
# For details: https://github.com/pylint-dev/pylint/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/pylint/blob/main/CONTRIBUTORS.txt

import os
import re
import sys
from pathlib import Path

import astroid
from astroid import nodes

from pylint.checkers import variables
from pylint.interfaces import HIGH
from pylint.testutils import CheckerTestCase, MessageTest, linter, set_config
from pylint.testutils.reporter_for_tests import GenericTestReporter

REGR_DATA_DIR = str(Path(__file__).parent / ".." / "regrtest_data")


class TestVariablesChecker(CheckerTestCase):
    CHECKER_CLASS = variables.VariablesChecker

    def test_all_elements_without_parent(self) -> None:
        node = astroid.extract_node("__all__ = []")
        node.value.elts.append(nodes.Const("test", parent=None))
        root = node.root()
        with self.assertNoMessages():
            self.checker.visit_module(root)
            self.checker.leave_module(root)


class TestVariablesCheckerWithTearDown(CheckerTestCase):
    CHECKER_CLASS = variables.VariablesChecker

    def setup_method(self) -> None:
        super().setup_method()
        self._to_consume_backup = self.checker._to_consume
        self.checker._to_consume = []

    def teardown_method(self) -> None:
        self.checker._to_consume = self._to_consume_backup

    @set_config(callbacks=("callback_", "_callback"))
    def test_custom_callback_string(self) -> None:
        """Test the --callbacks option works."""
        node = astroid.extract_node(
            """
        def callback_one(abc):
             ''' should not emit unused-argument. '''
        """
        )
        with self.assertNoMessages():
            self.checker.visit_functiondef(node)
            self.checker.leave_functiondef(node)

        node = astroid.extract_node(
            """
        def two_callback(abc, defg):
             ''' should not emit unused-argument. '''
        """
        )
        with self.assertNoMessages():
            self.checker.visit_functiondef(node)
            self.checker.leave_functiondef(node)

        node = astroid.extract_node(
            """
        def normal_func(abc):
             ''' should emit unused-argument. '''
        """
        )
        with self.assertAddsMessages(
            MessageTest(
                "unused-argument",
                node=node["abc"],
                args="abc",
                confidence=HIGH,
                line=2,
                col_offset=16,
                end_line=2,
                end_col_offset=19,
            )
        ):
            self.checker.visit_functiondef(node)
            self.checker.leave_functiondef(node)

        node = astroid.extract_node(
            """
        def cb_func(abc):
             ''' Previous callbacks are overridden. '''
        """
        )
        with self.assertAddsMessages(
            MessageTest(
                "unused-argument",
                node=node["abc"],
                args="abc",
                confidence=HIGH,
                line=2,
                col_offset=12,
                end_line=2,
                end_col_offset=15,
            )
        ):
            self.checker.visit_functiondef(node)
            self.checker.leave_functiondef(node)

    @set_config(redefining_builtins_modules=("os",))
    def test_redefined_builtin_modname_not_ignored(self) -> None:
        node = astroid.parse(
            """
        from future.builtins import open
        """
        )
        with self.assertAddsMessages(
            MessageTest(
                "redefined-builtin",
                node=node.body[0],
                args="open",
                line=2,
                col_offset=0,
                end_line=2,
                end_col_offset=32,
            )
        ):
            self.checker.visit_module(node)

    @set_config(redefining_builtins_modules=("os",))
    def test_redefined_builtin_in_function(self) -> None:
        node = astroid.extract_node(
            """
        def test():
            from os import open
        """
        )
        with self.assertNoMessages():
            self.checker.visit_module(node.root())
            self.checker.visit_functiondef(node)

    def test_import_as_underscore(self) -> None:
        node = astroid.parse(
            """
        import math as _
        """
        )
        with self.assertNoMessages():
            self.walk(node)

    def test_lambda_in_classdef(self) -> None:
        # Make sure lambda doesn't raises
        # Undefined-method in class def

        # Issue 1824
        # https://github.com/pylint-dev/pylint/issues/1824
        node = astroid.parse(
            """
        class MyObject(object):
            method1 = lambda func: func()
            method2 = lambda function: function()
        """
        )
        with self.assertNoMessages():
            self.walk(node)

    def test_nested_lambda(self) -> None:
        """Make sure variables from parent lambdas
        aren't noted as undefined.

        https://github.com/pylint-dev/pylint/issues/760
        """
        node = astroid.parse(
            """
        lambda x: lambda: x + 1
        """
        )
        with self.assertNoMessages():
            self.walk(node)

    @set_config(ignored_argument_names=re.compile("arg"))
    def test_ignored_argument_names_no_message(self) -> None:
        """Make sure is_ignored_argument_names properly ignores
        function arguments.
        """
        node = astroid.parse(
            """
        def fooby(arg):
            pass
        """
        )
        with self.assertNoMessages():
            self.walk(node)

    @set_config(ignored_argument_names=re.compile("args|kwargs"))
    def test_ignored_argument_names_starred_args(self) -> None:
        node = astroid.parse(
            """
        def fooby(*args, **kwargs):
            pass
        """
        )
        with self.assertNoMessages():
            self.walk(node)


class TestMissingSubmodule(CheckerTestCase):
    CHECKER_CLASS = variables.VariablesChecker

    @staticmethod
    def test_package_all() -> None:
        sys.path.insert(0, REGR_DATA_DIR)
        try:
            linter.check([os.path.join(REGR_DATA_DIR, "package_all")])
            assert isinstance(linter.reporter, GenericTestReporter)
            got = linter.reporter.finalize().strip()
            assert got == "E:  3: Undefined variable name 'missing' in __all__"
        finally:
            sys.path.pop(0)



# [line-too-long, missing-module-docstring, undefined-all-variable, undefined-all-variable, undefined-all-variable, undefined-all-variable, undefined-all-variable]
# ↓ equivalent to __all__ = ["C", "O", "N", "S", "T"]
__all__ = list("CONST")
#
CONST = 42



# pylint: disable=missing-docstring


from typing import TYPE_CHECKING


if TYPE_CHECKING:
    import math


VAR = 'pylint'  # [unused-variable]


try:
    pass
except TypeError as exc:
    print(exc)
except ValueError as exc:
    print(exc)



from typing import Optional, Union


def forecast(
    temp: Union[int, float],  # [consider-alternative-union-syntax]
    unit: Optional[str],  # [consider-alternative-union-syntax]
) -> None:
    print(f'Temperature: {temp}{unit or ""}')



"""Tests for used-before-assignment with python 3.10's pattern matching"""

match ("example", "one"):
    case (x, y) if x == "example":
        print("x used to cause used-before-assignment!")
    case _:
        print("good thing it doesn't now!")


# pylint: disable = missing-function-docstring, redefined-outer-name, missing-class-docstring

# https://github.com/pylint-dev/pylint/issues/9668
from enum import Enum
from pylint.constants import PY311_PLUS
if PY311_PLUS:
    from typing import assert_never  # pylint: disable=no-name-in-module
else:
    from typing_extensions import assert_never

class Example(Enum):
    FOO = 1
    BAR = 2

def check_value_if_then_match_return(example: Example, should_check: bool) -> str | None:
    if should_check:
        result = None
    else:
        match example:
            case Example.FOO:
                result = "foo"
            case Example.BAR:
                result = "bar"
            case _:
                return None

    return result

def check_value_if_then_match_raise(example: Example, should_check: bool) -> str | None:
    if should_check:
        result = None
    else:
        match example:
            case Example.FOO:
                result = "foo"
            case Example.BAR:
                result = "bar"
            case _:
                raise ValueError("Not a valid enum")

    return result

def check_value_if_then_match_assert_never(example: Example, should_check: bool) -> str | None:
    if should_check:
        result = None
    else:
        match example:
            case Example.FOO:
                result = "foo"
            case Example.BAR:
                result = "bar"
            case _:
                assert_never(example)

    return result

def g(x):
    if x is None:
        y = 0
    else:
        match x:
            case int():
                y = x
            case _:
                raise TypeError(type(x))

    return y

def check_value_if_then_match_nested(
    example: Example, example_inner: Example, should_check: bool
) -> str | None:
    if should_check:
        result = None
    else:
        match example:
            case Example.FOO:
                match example_inner:
                    case Example.BAR:
                        result = "bar"
                    case _:
                        return None
            case _:
                return None

    return result

def check_value_if_then_match_non_exhaustive(example: Example, should_check: bool) -> str | None:
    if should_check:
        result = None
    else:
        match example:
            case Example.FOO:
                result = "foo"
            case Example.BAR:
                pass
            case _:
                return None

    return result  # [possibly-used-before-assignment]



"""Tests for invalid name for names declared at module level"""
# pylint: disable=missing-class-docstring, too-few-public-methods, missing-function-docstring, wrong-import-position

import collections

Class = collections.namedtuple("a", ("b", "c"))


class ClassA:
    pass


ClassB = ClassA


def A():  # [invalid-name]
    return 1, 2, 3


CONSTA, CONSTB, CONSTC = A()
CONSTD = A()

CONST = "12 34 ".rstrip().split()


ASSIGNMENT_THAT_CRASHED_PYLINT = type(float.__new__.__code__)


# Exclusive assignment: uses const regex
if CONST:
    OTHER_CONST = 1
elif CONSTA:
    OTHER_CONST = 2
else:
    OTHER_CONST = 3


# Lists, sets, and objects can pass against the variable OR const regexes.
if CONST:
    other_const = [1]
elif CONSTA:
    other_const = [2]
else:
    other_const = [3]


from importlib.metadata import PackageNotFoundError
from importlib.metadata import version

try:
    VERSION = version("ty")  # uninferable
except PackageNotFoundError:
    VERSION = "0.0.0"



def calculate_drag_force(*, velocity, area, density, drag_coefficient):
    """This function is 'Keyword only' for all args due to the '*'."""
    return 0.5 * drag_coefficient * density * area * velocity**2


# This is now impossible to do and will raise a TypeError:
# drag_force = calculate_drag_force(30, 2.5, 1.225, 0.47)
#              ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
# TypeError: calculate_drag_force() takes 0 positional arguments but 4 were given

# And this is the only way to call 'calculate_drag_force'
drag_force = calculate_drag_force(
    velocity=30, area=2.5, density=1.225, drag_coefficient=0.47
)



# Licensed under the GPL: https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
# For details: https://github.com/pylint-dev/pylint/blob/main/LICENSE
# Copyright (c) https://github.com/pylint-dev/pylint/blob/main/CONTRIBUTORS.txt

"""Unit test for the extensions.diadefslib modules."""

# pylint: disable=redefined-outer-name

from __future__ import annotations

import re
from collections.abc import Callable, Iterator, Sequence
from pathlib import Path
from unittest.mock import Mock

import pytest
from astroid import extract_node, nodes

from pylint.pyreverse.diadefslib import (
    ClassDiadefGenerator,
    DefaultDiadefGenerator,
    DiaDefGenerator,
    DiadefsHandler,
)
from pylint.pyreverse.diagrams import DiagramEntity, Relationship
from pylint.pyreverse.inspector import Linker, Project
from pylint.testutils.pyreverse import PyreverseConfig
from pylint.testutils.utils import _test_cwd
from pylint.typing import GeneratorFactory, GetProjectCallable

HERE = Path(__file__)
TESTS = HERE.parent.parent


def _process_classes(classes: list[DiagramEntity]) -> list[tuple[bool, str]]:
    """Extract class names of a list."""
    return sorted((isinstance(c.node, nodes.ClassDef), c.title) for c in classes)


def _process_relations(
    relations: dict[str, list[Relationship]],
) -> list[tuple[str, str, str]]:
    """Extract relation indices from a relation list."""
    result = []
    for rel_type, rels in relations.items():
        for rel in rels:
            result.append((rel_type, rel.from_object.title, rel.to_object.title))
    result.sort()
    return result


@pytest.fixture
def HANDLER(
    default_config: PyreverseConfig, default_args: Sequence[str]
) -> DiadefsHandler:
    return DiadefsHandler(config=default_config, args=default_args)


@pytest.fixture(scope="module")
def PROJECT(get_project: GetProjectCallable) -> Iterator[Project]:
    with _test_cwd(TESTS):
        yield get_project("data")


# Fixture with factory function to create DiaDefGenerator instances
@pytest.fixture
def generator_factory(HANDLER: DiadefsHandler, PROJECT: Project) -> GeneratorFactory:
    def _factory(
        config: PyreverseConfig | None = None, args: Sequence[str] | None = None
    ) -> DiaDefGenerator:
        if config:
            HANDLER.config = config
        if args:
            HANDLER.args = args
        return DiaDefGenerator(Linker(PROJECT), HANDLER)

    return _factory


# Fixture for creating mocked nodes
@pytest.fixture
def mock_node() -> Callable[[str], Mock]:
    def _mock_node(module_path: str) -> Mock:
        """Create a mocked node with a given module path."""
        node = Mock()
        node.root.return_value.name = module_path
        return node

    return _mock_node


def test_option_values(
    default_config: PyreverseConfig,
    default_args: Sequence[str],
    HANDLER: DiadefsHandler,
    PROJECT: Project,
) -> None:
    """Test for ancestor, associated and module options."""
    df_h = DiaDefGenerator(Linker(PROJECT), HANDLER)
    cl_config = default_config
    cl_config.classes = ["Specialization"]
    cl_h = DiaDefGenerator(Linker(PROJECT), DiadefsHandler(config=cl_config, args=[]))
    assert df_h._get_levels() == (0, 0)
    assert not df_h.module_names
    assert cl_h._get_levels() == (-1, -1)
    assert cl_h.module_names
    for hndl in (df_h, cl_h):
        hndl.config.all_ancestors = True
        hndl.config.all_associated = True
        hndl.config.module_names = True
        hndl._set_default_options()
        assert hndl._get_levels() == (-1, -1)
        assert hndl.module_names
    handler = DiadefsHandler(config=default_config, args=default_args)
    df_h = DiaDefGenerator(Linker(PROJECT), handler)
    cl_config = default_config
    cl_config.classes = ["Specialization"]
    cl_h = DiaDefGenerator(
        Linker(PROJECT), DiadefsHandler(config=cl_config, args=default_args)
    )
    for hndl in (df_h, cl_h):
        hndl.config.show_ancestors = 2
        hndl.config.show_associated = 1
        hndl.config.module_names = False
        hndl._set_default_options()
        assert hndl._get_levels() == (2, 1)
        assert not hndl.module_names


def test_default_values() -> None:
    """Test default values for package or class diagrams."""
    # TODO : should test difference between default values for package or class diagrams


class TestShowOptions:
    def test_show_stdlib(self) -> None:
        example = extract_node(
            '''
            import collections

            class CustomDict(collections.OrderedDict):
                """docstring"""
            '''
        )

        config = PyreverseConfig()
        args: Sequence[str] = []
        dd_gen = DiaDefGenerator(Linker(PROJECT), DiadefsHandler(config, args))

        # Default behavior
        assert not list(dd_gen.get_ancestors(example, 1))

        # Show standard library enabled
        config.show_stdlib = True
        ancestors = list(dd_gen.get_ancestors(example, 1))
        assert len(ancestors) == 1
        assert ancestors[0].name == "OrderedDict"

    def test_show_builtin(self) -> None:
        example = extract_node(
            '''
            class CustomError(Exception):
                """docstring"""
            '''
        )

        config = PyreverseConfig()
        args: Sequence[str] = []
        dd_gen = DiaDefGenerator(Linker(PROJECT), DiadefsHandler(config, args))

        # Default behavior
        assert not list(dd_gen.get_ancestors(example, 1))

        # Show builtin enabled
        config.show_builtin = True
        ancestors = list(dd_gen.get_ancestors(example, 1))
        assert len(ancestors) == 1
        assert ancestors[0].name == "Exception"


class TestDefaultDiadefGenerator:
    _should_rels = [
        ("aggregation", "DoNothing2", "Specialization"),
        ("composition", "DoNothing", "Ancestor"),
        ("composition", "DoNothing", "Specialization"),
        ("specialization", "Specialization", "Ancestor"),
    ]

    def test_extract_relations(self, HANDLER: DiadefsHandler, PROJECT: Project) -> None:
        """Test extract_relations between classes."""
        cd = DefaultDiadefGenerator(Linker(PROJECT), HANDLER).visit(PROJECT)[1]
        cd.extract_relationships()
        relations = _process_relations(cd.relationships)
        assert relations == self._should_rels

    def test_functional_relation_extraction(
        self,
        default_config: PyreverseConfig,
        default_args: Sequence[str],
        get_project: GetProjectCallable,
    ) -> None:
        """Functional test of relations extraction;
        different classes possibly in different modules.
        """
        # XXX should be catching pyreverse environment problem but doesn't
        # pyreverse doesn't extract the relations but this test ok
        project = get_project("data")
        handler = DiadefsHandler(default_config, default_args)
        diadefs = handler.get_diadefs(project, Linker(project, tag=True))
        cd = diadefs[1]
        relations = _process_relations(cd.relationships)
        assert relations == self._should_rels


def test_known_values1(HANDLER: DiadefsHandler, PROJECT: Project) -> None:
    dd = DefaultDiadefGenerator(Linker(PROJECT), HANDLER).visit(PROJECT)
    assert len(dd) == 2
    keys = [d.TYPE for d in dd]
    assert keys == ["package", "class"]
    pd = dd[0]
    assert pd.title == "packages No Name"
    modules = sorted((isinstance(m.node, nodes.Module), m.title) for m in pd.objects)
    assert modules == [
        (True, "data"),
        (True, "data.clientmodule_test"),
        (True, "data.nullable_pattern"),
        (True, "data.property_pattern"),
        (True, "data.suppliermodule_test"),
    ]
    cd = dd[1]
    assert cd.title == "classes No Name"
    classes = _process_classes(cd.objects)
    assert classes == [
        (True, "Ancestor"),
        (True, "CustomException"),
        (True, "DoNothing"),
        (True, "DoNothing2"),
        (True, "DoSomething"),
        (True, "Interface"),
        (True, "NullablePatterns"),
        (True, "PropertyPatterns"),
        (True, "Specialization"),
    ]


def test_known_values2(
    HANDLER: DiadefsHandler, get_project: GetProjectCallable
) -> None:
    project = get_project("data.clientmodule_test")
    dd = DefaultDiadefGenerator(Linker(project), HANDLER).visit(project)
    assert len(dd) == 1
    keys = [d.TYPE for d in dd]
    assert keys == ["class"]
    cd = dd[0]
    assert cd.title == "classes No Name"
    classes = _process_classes(cd.objects)
    assert classes == [(True, "Ancestor"), (True, "Specialization")]


def test_known_values3(HANDLER: DiadefsHandler, PROJECT: Project) -> None:
    HANDLER.config.classes = ["Specialization"]
    cdg = ClassDiadefGenerator(Linker(PROJECT), HANDLER)
    special = "data.clientmodule_test.Specialization"
    cd = cdg.class_diagram(PROJECT, special)
    assert cd.title == special
    classes = _process_classes(cd.objects)
    assert classes == [
        (True, "data.clientmodule_test.Ancestor"),
        (True, special),
        (True, "data.suppliermodule_test.DoNothing"),
        (True, "data.suppliermodule_test.DoNothing2"),
    ]


def test_known_values4(HANDLER: DiadefsHandler, PROJECT: Project) -> None:
    HANDLER.config.classes = ["Specialization"]
    HANDLER.config.module_names = False
    cd = ClassDiadefGenerator(Linker(PROJECT), HANDLER).class_diagram(
        PROJECT, "data.clientmodule_test.Specialization"
    )
    assert cd.title == "data.clientmodule_test.Specialization"
    classes = _process_classes(cd.objects)
    assert classes == [
        (True, "Ancestor"),
        (True, "DoNothing"),
        (True, "DoNothing2"),
        (True, "Specialization"),
    ]


def test_regression_dataclasses_inference(
    HANDLER: DiadefsHandler, get_project: GetProjectCallable
) -> None:
    project_path = Path("regrtest_data") / "dataclasses_pyreverse"
    path = get_project(str(project_path))

    cdg = ClassDiadefGenerator(Linker(path), HANDLER)
    special = "regrtest_data.dataclasses_pyreverse.InventoryItem"
    cd = cdg.class_diagram(path, special)
    assert cd.title == special


def test_should_include_by_depth_no_limit(
    generator_factory: GeneratorFactory, mock_node: Mock
) -> None:
    """Test that nodes are included when no depth limit is set."""
    generator = generator_factory()

    # Create mocked nodes with different depths
    node1 = mock_node("pkg")  # Depth 0
    node2 = mock_node("pkg.subpkg")  # Depth 1
    node3 = mock_node("pkg.subpkg.module")  # Depth 2

    # All nodes should be included when max_depth is None
    assert generator._should_include_by_depth(node1)
    assert generator._should_include_by_depth(node2)
    assert generator._should_include_by_depth(node3)


@pytest.mark.parametrize("max_depth", range(5))
def test_should_include_by_depth_absolute(
    generator_factory: GeneratorFactory, mock_node: Mock, max_depth: int
) -> None:
    """Test absolute depth filtering when root package is specified.

    - 'pkg'                  -> depth 0
    - 'pkg.subpkg'           -> depth 1
    - 'pkg.subpkg.module'    -> depth 2
    - 'pkg.subpkg.module.submodule' -> depth 3
    """
    specified_pkg = ["pkg"]
    generator = generator_factory(PyreverseConfig(max_depth=max_depth), specified_pkg)

    test_cases = {
        "pkg": [True, True, True, True, True],
        "pkg.subpkg": [False, True, True, True, True],
        "pkg.subpkg.module": [False, False, True, True, True],
        "pkg.subpkg.module.submodule": [False, False, False, True, True],
    }
    nodes = [mock_node(path) for path, _ in test_cases.items()]

    for node in nodes:
        should_show = test_cases[node.root.return_value.name][max_depth]
        result = generator._should_include_by_depth(node)

        msg = (
            f"Node {node.root.return_value.name} with max_depth={max_depth} and "
            f"specified package {specified_pkg} should"
            f"{'' if should_show else ' not'} show. "
            f"Generator returns: {result}"
        )
        assert result == should_show, msg


@pytest.mark.parametrize("max_depth", range(5))
@pytest.mark.parametrize("args", [["pkg"], ["pkg.subpkg"], ["pkg.subpkg.module"]])
def test_should_include_by_depth_relative_single_package(
    generator_factory: GeneratorFactory,
    mock_node: Mock,
    max_depth: int,
    args: list[str],
) -> None:
    """Test relative depth filtering when only one package is specified.

    Each test case ensures that depth is calculated **relative** to the specified package.
    """
    specified_pkg = args[0]
    generator = generator_factory(PyreverseConfig(max_depth=max_depth), args)

    test_cases = {
        "pkg": {
            "pkg": [True, True, True, True, True],
            "pkg.subpkg": [False, True, True, True, True],
            "pkg.subpkg.module": [False, False, True, True, True],
            "pkg.subpkg.module.submodule": [False, False, False, True, True],
        },
        "pkg.subpkg": {
            "pkg": [False, False, False, False, False],
            "pkg.subpkg": [True, True, True, True, True],
            "pkg.subpkg.module": [False, True, True, True, True],
            "pkg.subpkg.module.submodule": [False, False, True, True, True],
        },
        "pkg.subpkg.module": {
            "pkg": [False, False, False, False, False],
            "pkg.subpkg": [False, False, False, False, False],
            "pkg.subpkg.module": [True, True, True, True, True],
            "pkg.subpkg.module.submodule": [False, True, True, True, True],
        },
    }
    nodes = [mock_node(path) for path, _ in test_cases.items()]

    for node in nodes:
        should_show = test_cases[specified_pkg][node.root.return_value.name][max_depth]
        result = generator._should_include_by_depth(node)

        msg = (
            f"Node {node.root.return_value.name} with max_depth={max_depth} and "
            f"specified package {specified_pkg} should"
            f"{'' if should_show else ' not'} show. "
            f"Generator returns: {result}"
        )
        assert result == should_show, msg


@pytest.mark.parametrize("max_depth", range(5))
def test_should_include_by_depth_relative_multiple_packages(
    generator_factory: GeneratorFactory,
    mock_node: Mock,
    max_depth: int,
) -> None:
    """Test relative depth filtering when multiple packages are specified."""
    specified_pkg = ["pkg", "pkg.subpkg"]

    warning_msg = re.escape(
        "Detected nested names within the specified packages. "
        "The following packages: ['pkg'] will be ignored for "
        "depth calculations, using only: ['pkg.subpkg'] as the base for limiting "
        "package depth."
    )

    with pytest.warns(UserWarning, match=warning_msg):
        generator = generator_factory(
            PyreverseConfig(max_depth=max_depth), specified_pkg
        )

    test_cases = {
        "pkg": [False, False, False, False, False],
        "pkg.subpkg": [True, True, True, True, True],
        "pkg.subpkg.module": [False, True, True, True, True],
        "pkg.subpkg.module.submodule": [False, False, True, True, True],
    }
    nodes = [mock_node(path) for path, _ in test_cases.items()]

    for node in nodes:
        should_show = test_cases[node.root.return_value.name][max_depth]
        result = generator._should_include_by_depth(node)

        msg = (
            f"Node {node.root.return_value.name} with max_depth={max_depth} and "
            f"specified package {specified_pkg} "
            f"{'should show' if should_show else 'should not show'}. "
            f"Generator returns: {result}"
        )
        assert result == should_show, msg


def test_get_leaf_nodes(generator_factory: GeneratorFactory) -> None:
    """Test that leaf nodes are correctly identified."""
    # Create a tree with a single root node and multiple leaf nodes
    specified_packeges = [
        "pkg",
        "pkg.subpkg1",
        "pkg.subpkg2",
        "pkg.subpkg1.module1",
        "pkg.subpkg2.module2",
        "pkg.subpkg1.module1.submodule",
    ]
    corr = set(["pkg.subpkg2.module2", "pkg.subpkg1.module1.submodule"])

    generator = generator_factory(args=specified_packeges)
    assert len(corr.difference(generator.get_leaf_nodes())) == 0



# pylint: disable=missing-function-docstring,missing-module-docstring,missing-class-docstring,too-few-public-methods
# pylint: disable=unused-argument

def f[T](a: T) -> T:
    print(a)

class SomeClass[T, *Ts, **P]:
    def __init__(self, value: T):
        self.value = value


class Parent[T]:
    def __init__(self, x: T, y: S) -> None:  # [undefined-variable]
        ...

class Child[T](Parent[T, S]):  # [undefined-variable]
    ...



"""Tests for fixme in docstrings"""
# pylint: disable=missing-function-docstring, pointless-string-statement

# +1: [fixme]
"""TODO resolve this"""
# +1: [fixme]
""" TODO: indentations are permitted """
# +1: [fixme]
''' TODO: indentations are permitted '''
# +1: [fixme]
"""    TODO: indentations are permitted"""

""" preceding text TODO: is not permitted"""

"""
FIXME don't forget this # [fixme]
XXX also remember this # [fixme]
FIXME: and this line, but treat it as one FIXME TODO # [fixme]
text cannot precede the TODO: it must be at the start
      XXX indentations are okay # [fixme]
??? the codetag must be recognized
"""

# +1: [fixme]
# FIXME should still work

# +1: [fixme]
# TODO """ should work

# """ TODO will not work
"""# TODO will not work"""

"""TODOist API should not result in a message"""

# +2: [fixme]
"""
TO make something DO: look a regex
"""

# pylint: disable-next=fixme
"""TODO won't work anymore"""

# +2: [fixme]
def function():
    """./TODO implement this"""


'''
  XXX single quotes should be no different # [fixme]
'''
def function2():
    '''
    ./TODO implement this # [fixme]
    FIXME and this # [fixme]
    '''
    '''FIXME one more for good measure''' # [fixme]



"""empty-comment test-case"""
# +1:[empty-comment]
a = 5  #
# +1:[empty-comment]
#
a = '#' + '1'
# +1:[empty-comment]
print(a)  #
print("A=", a)  # should not be an error#
# +1:[empty-comment]
a = "#pe\0ace#love#"  #
a = "peace#love"  # \0 peace'#'''' love#peace'''-'#love'-"peace#love"#
#######



def get_user_code(name):
    return input(f"Enter code to be executed please, {name}: ")


username = "Ada"
# If the globals dictionary does not contain a value for the key __builtins__,
# all builtins are allowed. You need to be explicit about it being disallowed.
allowed_globals = {"__builtins__": {}}
allowed_locals = {}
# pylint: disable-next=exec-used
exec(get_user_code(username), allowed_globals, allowed_locals)
